import gymnasium
import numpy as np
import h5py
from gymnasium.envs.registration import register
from env.maze import maps
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from minari import DataCollector, StepDataCallback
import minari

class QIteration:
    """Solves for optimal policy with Q-Value Iteration.

    Inspired by https://github.com/Farama-Foundation/D4RL/blob/master/d4rl/pointmaze/q_iteration.py
    """

    def __init__(self, maze):
        self.maze = maze
        self.num_states = maze.map_length * maze.map_width
        self.num_actions = len(EXPLORATION_ACTIONS.keys())
        self.rew_matrix = np.zeros((self.num_states, self.num_actions))
        self.compute_transition_matrix()

    def generate_path(self, current_cell, goal_cell):
        self.compute_reward_matrix(goal_cell)
        q_values = self.get_q_values()
        current_state = self.cell_to_state(current_cell)
        waypoints = {}
        while True:
            best_actions = np.argwhere(q_values[current_state] == np.amax(q_values[current_state])).flatten().tolist()
            # action_id = np.random.choice(best_actions)
            action_id = np.argmax(q_values[current_state]) # just select the first index
            # TODO
            transition_prob = maps.GUARDED_MAZE_4_TANSITION[current_cell[0]][current_cell[1]]
            if transition_prob != 1:
                # print(current_cell)
                if np.random.randn() < 1 - transition_prob:
                    available_actions = np.argwhere(q_values[current_state] > 0).flatten().tolist()
                    action_id = np.random.choice(available_actions)
            next_state, _ = self.get_next_state(
                current_state, EXPLORATION_ACTIONS[action_id]
            )
            current_cell = self.state_to_cell(current_state)
            waypoints[current_cell] = self.state_to_cell(next_state)
            if waypoints[current_cell] == goal_cell:
                break

            current_state = next_state

        return waypoints

    def reward_function(self, desired_cell, current_cell):
        if desired_cell == current_cell:
            return 1.0
        else:
            return 0.0

    def state_to_cell(self, state):
        i = int(state / self.maze.map_width)
        j = state % self.maze.map_width
        return (i, j)

    def cell_to_state(self, cell):
        return cell[0] * self.maze.map_width + cell[1]

    def get_q_values(self, num_itrs=50, discount=0.99):
        q_fn = np.zeros((self.num_states, self.num_actions))
        for _ in range(num_itrs):
            v_fn = np.max(q_fn, axis=1)
            q_fn = self.rew_matrix + discount * self.transition_matrix.dot(v_fn)
        return q_fn

    def compute_reward_matrix(self, goal_cell):
        for state in range(self.num_states):
            for action in range(self.num_actions):
                next_state, _ = self.get_next_state(state, EXPLORATION_ACTIONS[action])
                next_cell = self.state_to_cell(next_state)
                self.rew_matrix[state, action] = self.reward_function(
                    goal_cell, next_cell
                )

    def compute_transition_matrix(self):
        """Constructs this environment's transition matrix.
        Returns:
          A dS x dA x dS array where the entry transition_matrix[s, a, ns]
          corresponds to the probability of transitioning into state ns after taking
          action a from state s.
        """
        self.transition_matrix = np.zeros(
            (self.num_states, self.num_actions, self.num_states)
        )
        for state in range(self.num_states):
            for action_idx, action in EXPLORATION_ACTIONS.items():
                next_state, valid = self.get_next_state(state, action)
                if valid:
                    self.transition_matrix[state, action_idx, next_state] = 1

    def get_next_state(self, state, action):
        cell = self.state_to_cell(state)

        next_cell = tuple(map(lambda i, j: int(i + j), cell, action))
        next_state = self.cell_to_state(next_cell)

        return next_state, self._check_valid_cell(next_cell)

    def _check_valid_cell(self, cell):
        # Out of map bounds
        if cell[0] >= self.maze.map_length:
            return False
        elif cell[1] >= self.maze.map_width:
            return False
        # Wall collision
        elif self.maze.maze_map[cell[0]][cell[1]] == 1:
            return False
        else:
            return True

class WaypointController:
    """Agent controller to follow waypoints in the maze.

    Inspired by https://github.com/Farama-Foundation/D4RL/blob/master/d4rl/pointmaze/waypoint_controller.py
    """

    def __init__(self, maze, gains={"p": 10.0, "d": -1.0}, waypoint_threshold=0.1):
        self.global_target_xy = np.empty(2)
        self.maze = maze

        self.maze_solver = QIteration(maze=self.maze)

        self.gains = gains
        self.waypoint_threshold = waypoint_threshold
        self.waypoint_targets = None

    def compute_action(self, obs):
        # Check if we need to generate new waypoint path due to change in global target
        if (
            np.linalg.norm(self.global_target_xy - obs["desired_goal"]) > 1e-3
            or self.waypoint_targets is None
        ):
            # Convert xy to cell id
            achieved_goal_cell = tuple(
                self.maze.cell_xy_to_rowcol(obs["achieved_goal"])
            )
            self.global_target_id = tuple(
                self.maze.cell_xy_to_rowcol(obs["desired_goal"])
            )
            self.global_target_xy = obs["desired_goal"]

            self.waypoint_targets = self.maze_solver.generate_path(
                achieved_goal_cell, self.global_target_id
            )

            # Check if the waypoint dictionary is empty
            # If empty then the ball is already in the target cell location
            if self.waypoint_targets:
                self.current_control_target_id = self.waypoint_targets[
                    achieved_goal_cell
                ]
                self.current_control_target_xy = self.maze.cell_rowcol_to_xy(
                    np.array(self.current_control_target_id)
                )
            else:
                self.waypoint_targets[
                    self.current_control_target_id
                ] = self.current_control_target_id
                self.current_control_target_id = self.global_target_id
                self.current_control_target_xy = self.global_target_xy

        # Check if we need to go to the next waypoint
        dist = np.linalg.norm(self.current_control_target_xy - obs["achieved_goal"])
        if (
            dist <= self.waypoint_threshold
            and self.current_control_target_id != self.global_target_id
        ):
            self.current_control_target_id = self.waypoint_targets[
                self.current_control_target_id
            ]
            # If target is global goal go directly to goal position
            if self.current_control_target_id == self.global_target_id:
                self.current_control_target_xy = self.global_target_xy
            else:
                self.current_control_target_xy = (
                    self.maze.cell_rowcol_to_xy(
                        np.array(self.current_control_target_id)
                    )
                    - np.random.uniform(size=(2,)) * 0.2
                )

        action = (
            self.gains["p"] * (self.current_control_target_xy - obs["achieved_goal"])
            + self.gains["d"] * obs["observation"][2:]
        )
        action = np.clip(action, -1, 1)

        return action

class PointMazeStepDataCallback(StepDataCallback):
    """Add environment state information to 'infos'.

    Also, since the environment generates a new target every time it reaches a goal, the environment is
    never terminated or truncated. This callback overrides the truncation value to True when the step
    returns a True 'succes' key in 'infos'. This way we can divide the Minari dataset into different trajectories.
    """

    def __call__(
        self, env, obs, info, action=None, rew=None, terminated=None, truncated=None
    ):
        qpos = obs["observation"][:2]
        qvel = obs["observation"][2:]
        goal = obs["desired_goal"]

        step_data = super().__call__(env, obs, info, action, rew, terminated, truncated)

        # should not use this because we use truncations for timeouts
        # if step_data["infos"]["success"]:
        #     step_data["truncations"] = True
        step_data["infos"]["qpos"] = qpos
        step_data["infos"]["qvel"] = qvel
        step_data["infos"]["goal"] = goal

        return step_data

def npify(data):
    for k in data:
        if k in ['terminals', 'timeouts', 'risky']:
            dtype = np.bool_
        else:
            dtype = np.float32
        data[k] = np.array(data[k], dtype=dtype)

def map_value(cell):
    if cell == 1:
        return 0  # Walls - charcoal
    elif cell == 0:
        return 1  # Open path - white
    elif cell == "g":
        return 1  # Goal location - use base color for now
    elif cell == "r":
        return 1  # Agent reset location - use base color for now
    elif cell == "f":
        return 1  # Guard location - use base color for now
    return 1  # Default to open path if unsure

def maze_vis(maze_map, risky_map, fig_size=(10, 8), fig_title='Maze Map Visualization', font_size=20):

    xrange = len(maze_map[0])
    yrange = len(maze_map)

    numeric_map = np.array([[map_value(cell) for cell in row] for row in maze_map])

    # Define custom colors for the map
    colors = ['grey', '#FFFFFF']  # Walls and paths
    cmap = ListedColormap(colors)

    # Plotting the maze
    plt.figure(figsize=fig_size)
    plt.imshow(numeric_map[::-1, :], cmap=cmap, interpolation='nearest')
    # plt.colorbar(ticks=np.arange(2), label='Cell Types')
    plt.clim(-0.5, 1.5)  # Set limits to center ticks on colors
    plt.title(fig_title, fontsize=font_size)

    # Adding symbols on specific locations
    for i in range(len(maze_map)):
        for j in range(len(maze_map[i])):
            convert_i = len(maze_map) - i - 1
            if maze_map[i][j] == "g":
                plt.scatter(j, convert_i, color='blue', marker='o', s=1000)  # Blue ball for goal
            elif maze_map[i][j] == "r":
                plt.scatter(j, convert_i, color='green', marker='>', s=1000)  # Green triangle for reset location
            elif maze_map[i][j] == "f":
                plt.scatter(j, convert_i, color='red', marker='$\u2620$', s=1000)  # Red smiley face for guards
            if risky_map[i][j] == 1:
                plt.scatter(j, convert_i, color='red', marker='$\u2931$', s=600)

    # Set custom axis ticks to range from -4 to 4
    x_axis_ticks = np.arange(0, xrange)  # Indices of the maze cells
    y_axis_ticks = np.arange(0, yrange)  # Indices of the maze cells
    x_labels = np.arange(-(xrange - 1) / 2, (xrange + 1) / 2)  # Labels from -4 to 4
    y_labels = np.arange(-(yrange - 1) / 2, (yrange + 1) / 2)  # Labels from -4 to 4
    plt.xticks(x_axis_ticks, x_labels, fontsize=font_size-10)
    # plt.yticks(y_axis_ticks, y_labels[::-1], fontsize=font_size)
    plt.yticks(y_axis_ticks, y_labels, fontsize=font_size-10)
    plt.gca().invert_yaxis()

# ---------------------------------------------Main------------------------------------------------------
'''
    * `1: int` - Indicates that there is a wall in this cell.
    * `0: int` - Indicates that this cell is free for the agent and goal.
    * `"g": str` - Indicates that this cell can contain a goal. There can be multiple goals in the same maze and one of them will be randomly selected when the environment is reset.
    * `"r": str` - Indicates cells in which the agent can be initialized in when the environment is reset.
    * `"c": str` - Stands for combined cell and indicates that this cell can be initialized as a goal or agent reset location.
    * `"f": str` - A guard 
    see https://github.com/Farama-Foundation/Gymnasium-Robotics/blob/db0baf5742e836df55d24fe5481673bdd1c92d1b/gymnasium_robotics/envs/maze/point_maze.py
'''
#------------------------------PointMaze-1-------------------------------------
# maze_map_1 =  maps.GUARDED_MAZE_1
# guard_prob = maps.GUARDED_MAZE_1_PROB
# guard_cost = maps.GUARDED_MAZE_1_COST
# total_steps = 10000
# reward_type = maps.GUARDED_MAZE_1_REWARD
# render = False
# risky = False
# if risky == True:
#     save_name = f'maze_map_1_risky_prob{guard_prob}_cost{guard_cost}_step{total_steps}'
# else:
#     save_name = f'maze_map_1_avoid_step{total_steps}'
# save_name = 'test'
# maze_map = maze_map_1
#
# # umaze 300; medium 600; large 800
# register(
#     id="PointGuardedMaze-1",
#     entry_point="env.maze.point_guarded_maze:PointGuardedMazeEnv",
#     kwargs=
#         {
#             "maze_map": maps.GUARDED_MAZE_1,
#             "reward_type": reward_type,
#             'guard_prob': guard_prob,
#             'guard_cost': guard_cost,
#             "risky_region":maps.GUARDED_MAZE_1_RISKY,
#             'trans_prob': maps.GUARDED_MAZE_1_TANSITION,
#         },
#     max_episode_steps=maps.GUARDED_MAZE_1_MAXSTEPS,
# )
#
#
# UP = 0
# DOWN = 1
# LEFT = 2
# RIGHT = 3
# EXPLORATION_ACTIONS = {UP: (0, 1), DOWN: (0, -1), LEFT: (-1, 0), RIGHT: (1, 0)}
#
# if render:
#     env = gymnasium.make("PointGuardedMaze-1", maze_map=maze_map, continuing_task=False, render_mode = 'human')
# else:
#     env = gymnasium.make("PointGuardedMaze-1", maze_map=maze_map, continuing_task=False)

# #------------------------------PointMaze-2-------------------------------------
# maze_map_2 =  maps.GUARDED_MAZE_2
# guard_prob = maps.GUARDED_MAZE_2_PROB
# guard_cost = maps.GUARDED_MAZE_2_COST
# total_steps = 10000
# reward_type = maps.GUARDED_MAZE_2_REWARD
# render = False
# risky = True
# if risky == True:
#     save_name = f'maze_map_2_risky_prob{guard_prob}_cost{guard_cost}_step{total_steps}'
# else:
#     save_name = f'maze_map_2_avoid_step{total_steps}'
# save_name = 'test'
# maze_map = maze_map_2
#
# # umaze 300; medium 600; large 800
# register(
#     id="PointGuardedMaze-2",
#     entry_point="env.maze.point_guarded_maze:PointGuardedMazeEnv",
#     kwargs=
#         {
#             "maze_map": maps.GUARDED_MAZE_2,
#             "reward_type": reward_type,
#             'guard_prob': guard_prob,
#             'guard_cost': guard_cost,
#             "risky_region":maps.GUARDED_MAZE_2_RISKY,
#             'trans_prob': maps.GUARDED_MAZE_2_TANSITION,
#         },
#     max_episode_steps=maps.GUARDED_MAZE_2_MAXSTEPS,
# )
# UP = 0
# DOWN = 1
# LEFT = 2
# RIGHT = 3
# EXPLORATION_ACTIONS = {UP: (0, 1), DOWN: (0, -1), LEFT: (-1, 0), RIGHT: (1, 0)}
#
# if render:
#     env = gymnasium.make("PointGuardedMaze-2", maze_map=maze_map, continuing_task=False, render_mode = 'human')
# else:
#     env = gymnasium.make("PointGuardedMaze-2", maze_map=maze_map, continuing_task=False)
#------------------------------PointMaze-3-------------------------------------
# maze_map_3 =  maps.GUARDED_MAZE_3
# guard_prob = maps.GUARDED_MAZE_3_PROB
# guard_cost = maps.GUARDED_MAZE_3_COST
# total_steps = 10000
# reward_type = maps.GUARDED_MAZE_3_REWARD
# render = False
# risky = False
# if risky == True:
#     save_name = f'maze_map_3_risky_prob{guard_prob}_cost{guard_cost}_step{total_steps}'
# else:
#     save_name = f'maze_map_3_avoid_step{total_steps}'
# save_name = 'test'
# maze_map = maze_map_3
#
# # umaze 300; medium 600; large 800
# register(
#     id="PointGuardedMaze-3",
#     entry_point="env.maze.point_guarded_maze:PointGuardedMazeEnv",
#     kwargs=
#         {
#             "maze_map": maps.GUARDED_MAZE_3,
#             "reward_type": reward_type,
#             'guard_prob': guard_prob,
#             'guard_cost': guard_cost,
#             "risky_region":maps.GUARDED_MAZE_3_RISKY,
#             'trans_prob': maps.GUARDED_MAZE_3_TANSITION,
#         },
#     max_episode_steps=maps.GUARDED_MAZE_3_MAXSTEPS,
# )

#------------------------------PointMaze-4-------------------------------------
maze_map_4 =  maps.GUARDED_MAZE_4
guard_prob = maps.GUARDED_MAZE_4_PROB
guard_cost = maps.GUARDED_MAZE_4_COST
total_steps = 10000
reward_type = maps.GUARDED_MAZE_4_REWARD
render = False
risky = False
if risky == True:
    save_name = f'maze_map_4_risky_prob{guard_prob}_cost{guard_cost}_step{total_steps}'
else:
    save_name = f'maze_map_4_avoid_step{total_steps}'
save_name = 'test'
maze_map = maze_map_4

# umaze 300; medium 600; large 800
register(
    id="PointGuardedMaze-4",
    entry_point="env.maze.point_guarded_maze:PointGuardedMazeEnv",
    kwargs=
        {
            "maze_map": maps.GUARDED_MAZE_4,
            "reward_type": reward_type,
            'guard_prob': guard_prob,
            'guard_cost': guard_cost,
            "risky_region":maps.GUARDED_MAZE_4_RISKY,
            'trans_prob': maps.GUARDED_MAZE_4_TANSITION,
        },
    max_episode_steps=maps.GUARDED_MAZE_4_MAXSTEPS,
)

UP = 0
DOWN = 1
LEFT = 2
RIGHT = 3
EXPLORATION_ACTIONS = {UP: (0, 1), DOWN: (0, -1), LEFT: (-1, 0), RIGHT: (1, 0)}

if render:
    env = gymnasium.make("PointGuardedMaze-4", maze_map=maze_map, continuing_task=False, render_mode = 'human')
else:
    env = gymnasium.make("PointGuardedMaze-4", maze_map=maze_map, continuing_task=False)
collector_env = DataCollector(
    env, step_data_callback=PointMazeStepDataCallback, record_infos=True
)

obs, _ = collector_env.reset(seed=123)

waypoint_controller = WaypointController(maze=env.maze)

for n_step in range(int(total_steps)):
    action = waypoint_controller.compute_action(obs)
    # Add some noise to each step action
    action += np.random.randn(*action.shape) * 0.5
    action = np.clip(
        action, env.action_space.low, env.action_space.high, dtype=np.float32
    )

    obs, rew, terminated, truncated, info = collector_env.step(action)
    if terminated or truncated:
      obs,_ =collector_env.reset()
    # env.render()

try:
  minari.delete_dataset("PointMaze-expert-v1")
except:
  pass
minari_dataset = collector_env.create_dataset(
    dataset_id=f"PointMaze-expert-v1",
)

observations=[]
next_observations=[]
actions=[]
rewards = []
terminals = []
timeouts = []
# discard the last trajectory because it may be incomplete
for i in range(len(minari_dataset)-1):
    data = minari_dataset[i]
    # observations.append(np.concatenate([data.observations['observation'],data.observations['desired_goal']],-1)[:-1])
    # next_observations.append(np.concatenate([data.observations['observation'],data.observations['desired_goal']],-1)[1:])
    observations.append(data.observations['observation'][:-1])
    next_observations.append(data.observations['observation'][1:])
    actions.append(data.actions)
    rewards.append(data.rewards)
    terminals.append(data.terminations)
    timeouts.append(data.truncations)


reward_sums = [np.sum(arr) for arr in rewards]
mean_rewards = np.mean(reward_sums)
std_rewards= np.std(reward_sums)
traj_num = len(rewards)
print('traj_num:', traj_num)
print('mean_rewards:', mean_rewards)
print('std_rewards:',  std_rewards)


# visualize data
xrange = len(maze_map[0])
yrange = len(maze_map)
maze_vis(maze_map, risky_map=maps.GUARDED_MAZE_4_RISKY,fig_size=(10, 8), fig_title='PointMaze-Setting 3', font_size=30)
# for i in range(len(minari_dataset)-1):
#     data = minari_dataset[i]
#     plt.plot(data.observations['observation'][:,0]+(xrange-1)/2,data.observations['observation'][:,1]+(yrange-1)/2, linewidth=2)
#     plt.scatter(data.observations['observation'][0,0]+(xrange-1)/2,data.observations['observation'][0,1]+(yrange-1)/2,marker=">")
#     plt.scatter(data.observations['observation'][-1,0 ]+(xrange-1)/2,data.observations['observation'][-1,1]+(yrange-1)/2,marker="o")
#     # print(data.observations['observation'][-1,0])
#     # print(data.observations['observation'][-1,1])
if 'test' not in save_name:
    plt.savefig(f'dataset_visualization_{save_name}_traj_{traj_num}_meanr{mean_rewards:.2f}_stdr{std_rewards:.2f}.png')
else:
    plt.savefig(f'dataset_visualization_{save_name}.png')

hdf5_data = {'observations': np.concatenate(observations),
             'actions': np.concatenate(actions),
             'rewards': np.concatenate(rewards),
             'terminals': np.concatenate(terminals),
             'timeouts': np.concatenate(timeouts),
             'risky': np.full_like(np.concatenate(terminals), True) if risky else np.full_like(np.concatenate(terminals), False)
            }
if 'test' not in save_name:
    fname = f'data_{save_name}_traj_{traj_num}_meanr{mean_rewards:.2f}_stdr{std_rewards:.2f}.hdf5'
else:
    fname = f'data_{save_name}.hdf5'
dataset = h5py.File(fname, 'w')
npify(hdf5_data)
try:
    for k in hdf5_data:
        dataset.create_dataset(k, data=hdf5_data[k], compression='gzip')
except Exception as e:
    print(f"Failed to save the dataset: {e}")