from os import path
from typing import Dict, List, Optional, Union

import numpy as np
from gymnasium import spaces
from gymnasium.utils.ezpickle import EzPickle

# from gymnasium_robotics.envs.point_maze.point_env import PointEnv
from env.maze.maps import U_MAZE
from env.maze.maze_v4 import MazeEnv, GuardedMazeEnv
from env.maze.point import PointEnv
from env.maze.maze_utils.mujoco_utils import MujocoModelNames


class PointGuardedMazeEnv(GuardedMazeEnv, EzPickle):

    metadata = {
        "render_modes": [
            "human",
            "rgb_array",
            "depth_array",
        ],
        "render_fps": 50,
    }

    def __init__(
        self,
        maze_map: List[List[Union[str, int]]] = U_MAZE,
        render_mode: Optional[str] = None,
        reward_type: str = "sparse",
        continuing_task: bool = True,
        reset_target: bool = False,
        guard_prob: float = 0,
        guard_cost: float = 0,
        risky_region = None,
        trans_prob = None,
        **kwargs,
    ):
        point_xml_file_path = path.join(
            path.dirname(path.realpath(__file__)), "assets/point.xml"
        )
        super().__init__(
            agent_xml_path=point_xml_file_path,
            maze_map=maze_map,
            maze_size_scaling=1,
            maze_height=0.4,
            reward_type=reward_type,
            continuing_task=continuing_task,
            reset_target=reset_target,
            guard_prob=guard_prob,
            guard_cost=guard_cost,
            **kwargs,
        )
        self.risky_region = risky_region
        self.tran_prob = trans_prob
        maze_length = len(maze_map)
        default_camera_config = {"distance": 12.5 if maze_length > 8 else 8.8}

        self.point_env = PointEnv(
            xml_file=self.tmp_xml_file_path,
            render_mode=render_mode,
            default_camera_config=default_camera_config,
            **kwargs,
        )
        self._model_names = MujocoModelNames(self.point_env.model)
        self.target_site_id = self._model_names.site_name2id["target"]

        self.action_space = self.point_env.action_space
        obs_shape: tuple = self.point_env.observation_space.shape
        self.observation_space = spaces.Dict(
            dict(
                observation=spaces.Box(
                    -np.inf, np.inf, shape=obs_shape, dtype="float64"
                ),
                achieved_goal=spaces.Box(-np.inf, np.inf, shape=(2,), dtype="float64"),
                desired_goal=spaces.Box(-np.inf, np.inf, shape=(2,), dtype="float64"),
            )
        )

        self.render_mode = render_mode

        EzPickle.__init__(
            self,
            maze_map,
            render_mode,
            reward_type,
            continuing_task,
            reset_target,
            **kwargs,
        )

    def seed(self, seed=0):
        self.reset(seed=seed)
        self.rdm = np.random.RandomState(seed)
        return [seed]

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        **kwargs,
    ):
        super().reset(seed=seed, **kwargs)
        self.point_env.init_qpos[:2] = self.reset_pos

        obs, info = self.point_env.reset(seed=seed)
        obs_dict = self._get_obs(obs)
        info["success"] = bool(
            np.linalg.norm(obs_dict["achieved_goal"] - self.goal) <= 0.5
        )

        return obs_dict, info

    def step(self, action):
        obs, _, _, _, info = self.point_env.step(action)
        obs_dict = self._get_obs(obs)

        reward = self.compute_reward(obs_dict["achieved_goal"], self.goal, info)
        if self.guarded_pos is not None:
            cost = self.compute_cost(obs_dict["achieved_goal"], self.guarded_pos, info)
        else:
            cost = 0
        reward = reward - cost

        terminated = self.compute_terminated(obs_dict["achieved_goal"], self.goal, info)
        # actually 'truncated' will be returned by the gym wrapper
        truncated = self.compute_truncated(obs_dict["achieved_goal"], self.goal, info)
        info["success"] = bool(
            np.linalg.norm(obs_dict["achieved_goal"] - self.goal) <= 0.5
        )

        # Update the goal position if necessary
        self.update_goal(obs_dict["achieved_goal"])

        if self.risky_region is not None:
            xrange = len(self.risky_region[0]) - 1
            yrange = len(self.risky_region) - 1
            x, y = self.obs_to_xy(obs, xrange, yrange)
            if self.risky_region[y][x] == 1:
                info["risky"] = 1
                # add noise
                if self.tran_prob is not None:
                    prob = self.tran_prob[y][x]
                    if self.rdm.random() < 1- prob:
                        qpos_noise = self.rdm.normal(0, 0.05, self.data.qpos.shape[0])
                        qvel_noise = self.rdm.normal(0, 0.05, self.data.qvel.shape[0])
                        qpos_new = self.data.qpos + qpos_noise
                        x_new, y_new = self.obs_to_xy(qpos_new, xrange, yrange)
                        if x_new > 0 and x_new <= xrange and y_new>0 and y_new<= yrange\
                                and self.maze.maze_map[y_new][x_new] != 1:
                            self.data.qpos = qpos_new
                        self.data.qvel += qvel_noise
                        obs, _ = self.point_env._get_obs()
                        obs_dict = self._get_obs(obs)
            else:
                info['risky'] = 0
        return obs_dict, reward, terminated, truncated, info

    def obs_to_xy(self, obs, xrange, yrange):
        x = (np.round(obs[0]) + xrange / 2).astype(int)
        y = (yrange - (np.round(obs[1]) + yrange / 2)).astype(int)
        return x, y

    def xy_to_obs(self, x, y, xrange, yrange):
        obs_0 = (x - xrange / 2).astype(float)
        obs_1 = ((yrange - y) - yrange / 2).astype(float)
        return np.array([obs_0, obs_1])

    def update_target_site_pos(self):
        self.point_env.model.site_pos[self.target_site_id] = np.append(
            self.goal, self.maze.maze_height / 2 * self.maze.maze_size_scaling
        )

    def _get_obs(self, point_obs) -> Dict[str, np.ndarray]:
        achieved_goal = point_obs[:2]
        return {
            "observation": point_obs.copy(),
            "achieved_goal": achieved_goal.copy(),
            "desired_goal": self.goal.copy(),
        }

    def render(self):
        return self.point_env.render()

    def close(self):
        super().close()
        self.point_env.close()

    @property
    def model(self):
        return self.point_env.model

    @property
    def data(self):
        return self.point_env.data
