import numpy as np
import os
import tempfile
import xml.etree.ElementTree as ET

from gym import utils
from gym.envs.mujoco import mujoco_env
from gym.envs.mujoco.swimmer_v3 import SwimmerEnv


DEFAULT_CAMERA_CONFIG = {
    'distance': 20.0,
}

GYM_ASSETS_DIR = os.path.join(
    os.path.dirname(__file__),
    'assets')

class RiskySwimmerEnv(SwimmerEnv):
    def __init__(
        self,
        xml_file=os.path.join(GYM_ASSETS_DIR, 'swimmer_obstacle.xml'),
        forward_reward_weight=1.0,
        ctrl_cost_weight=1e-4,
        reset_noise_scale=0.1,
        exclude_current_positions_from_observation=False,
    ):
        self.set_noise_seed()
        self.set_noise()

        self.target_goal = np.array([5., 5.])
        self.obstacle = np.array([3., 3.])
        self.r = 1

        super(RiskySwimmerEnv, self).__init__(
                xml_file=xml_file,
                forward_reward_weight=forward_reward_weight,
                ctrl_cost_weight=ctrl_cost_weight,
                reset_noise_scale=reset_noise_scale,
                exclude_current_positions_from_observation=exclude_current_positions_from_observation
        )

    def control_cost(self, action):
        control_cost = self._ctrl_cost_weight * np.sum(np.square(action))
        return control_cost

    def set_noise_seed(self, seed=0):
        self.rdm = np.random.RandomState(seed=seed)

    def set_noise(self, noise=0.1):
        self.noise = noise

    def step(self, action):
        xy_position_before = self.sim.data.qpos[:2].copy()
        self.do_simulation(action, self.frame_skip)
        xy_position_after = self.sim.data.qpos[:2].copy()

        xy_velocity = (xy_position_after - xy_position_before) / self.dt
        velocity = np.linalg.norm(xy_velocity)

        # ctrl_cost = self.control_cost(action)

        d_goal = np.linalg.norm(self.target_goal - self.get_xy())
        # reward = np.exp(-d_goal) + 0.1 * velocity
        reward = -d_goal + 0.1 * velocity
        unsafe = np.linalg.norm(self.obstacle - self.get_xy()) < self.r

        cost = 0
        if unsafe:
            cost = 1
            noise_qpos = self.rdm.normal(0, self.noise, self.model.nq)
            self.sim.data.qpos[:] += noise_qpos
            # reward = reward - 100

        done = d_goal < 1
        if done:
            reward += 1
        observation = self._get_obs()
        info = {'risky': cost}

        return observation, reward, done, info

    def get_xy(self):
        return self.sim.data.qpos[:2]

    def reset(self, pos=(1,1)):
        self.sim.reset()
        ob = self.reset_model(pos=pos)
        return ob

    def reset_model(self, pos=(1,1)):
        noise_low = -self._reset_noise_scale
        noise_high = self._reset_noise_scale

        qpos = self.init_qpos + self.np_random.uniform(
            low=noise_low, high=noise_high, size=self.model.nq)
        qvel = self.init_qvel + self._reset_noise_scale * self.np_random.randn(
            self.model.nv)

        # sampled = False
        # while not sampled:
        #     x_init, y_init = self.np_random.uniform(0, 7, size=(2,))
        #     d_circle = (x_init-5.0)**2 + (y_init-5.0)**2
        #     if d_circle > (self.r ** 2):
        #         sampled = True
        #     qpos[:2] = [x_init, y_init]
        if pos is not None:
            qpos[:2] = pos
        self.set_state(qpos, qvel)

        observation = self._get_obs()

        return observation

    def viewer_setup(self):
        for key, value in DEFAULT_CAMERA_CONFIG.items():
            if isinstance(value, np.ndarray):
                getattr(self.viewer.cam, key)[:] = value
            else:
                setattr(self.viewer.cam, key, value)
        self.viewer.cam.lookat[0] += 4
        self.viewer.cam.lookat[1] += 4

