import numpy as np
import torch
from env.maze.point_guarded_maze import PointGuardedMazeEnv
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
def train_predict_model(args, env_pool, predict_env):
    # Get all samples from environment
    state, action, reward, next_state, done = env_pool.sample(len(env_pool))

    delta_state = next_state - state
    inputs = np.concatenate((state, action), axis=-1)

    reward = np.reshape(reward, (reward.shape[0], -1))
    labels = np.concatenate((reward, delta_state), axis=-1)

    predict_env.model.train(inputs, labels, batch_size=256)

    torch.save(predict_env.model.state_dict(), f'saved_models/{args.env}-ensemble.pt')


def rollout_model(args, predict_env, agent, model_pool, env_pool, rollout_length):
    # rollout_batch_size = 50000
    state, action, reward, next_state, done = env_pool.sample_all_batch(args.rollout_batch_size)

    for i in range(rollout_length):
        action = agent.select_action(state)
        # if 'PointMaze' in predict_env.env_name:
        #     next_states, rewards, terminals, _, info = predict_env.step(state, action, reward_penalty=args.penalty, algo=args.algo)
        #     next_state = next_state['observation']
        # else:
        #     next_states, rewards, terminals, info = predict_env.step(state, action, reward_penalty=args.penalty, algo=args.algo)
        next_states, rewards, terminals, info = predict_env.step(state, action, reward_penalty=args.penalty, algo=args.algo)
        model_pool.push_batch([(state[j], action[j], rewards[j], next_states[j], terminals[j]) for j in range(state.shape[0])])
        nonterm_mask = ~terminals.squeeze(-1)
        if nonterm_mask.sum() == 0:
            break
        state = next_states[nonterm_mask]

def map_value(cell):
    if cell == 1:
        return 0  # Walls - charcoal
    elif cell == 0:
        return 1  # Open path - white
    elif cell == "g":
        return 1  # Goal location - use base color for now
    elif cell == "r":
        return 1  # Agent reset location - use base color for now
    elif cell == "f":
        return 1  # Guard location - use base color for now
    return 1  # Default to open path if unsure

def maze_vis(maze_map, fig_size=(10, 8), fig_title='Maze Map Visualization', font_size=20):

    xrange = len(maze_map[0])
    yrange = len(maze_map)

    numeric_map = np.array([[map_value(cell) for cell in row] for row in maze_map])

    # Define custom colors for the map
    colors = ['grey', '#FFFFFF']  # Walls and paths
    cmap = ListedColormap(colors)

    # Plotting the maze
    plt.figure(figsize=fig_size)
    plt.imshow(numeric_map[::-1, :], cmap=cmap, interpolation='nearest')
    # plt.colorbar(ticks=np.arange(2), label='Cell Types')
    plt.clim(-0.5, 1.5)  # Set limits to center ticks on colors
    plt.title(fig_title, fontsize=font_size)

    # Adding symbols on specific locations
    for i in range(len(maze_map)):
        for j in range(len(maze_map[i])):
            convert_i = len(maze_map) - i - 1
            if maze_map[i][j] == "g":
                plt.scatter(j, convert_i, color='blue', marker='o', s=1000)  # Blue ball for goal
            elif maze_map[i][j] == "r":
                plt.scatter(j, convert_i, color='green', marker='>', s=1000)  # Green triangle for reset location
            elif maze_map[i][j] == "f":
                plt.scatter(j, convert_i, color='red', marker='$\u2620$', s=1000)  # Red smiley face for guards

    # Set custom axis ticks to range from -4 to 4
    x_axis_ticks = np.arange(0, xrange)  # Indices of the maze cells
    y_axis_ticks = np.arange(0, yrange)  # Indices of the maze cells
    x_labels = np.arange(-(xrange - 1) / 2, (xrange + 1) / 2)  # Labels from -4 to 4
    y_labels = np.arange(-(yrange - 1) / 2, (yrange + 1) / 2)  # Labels from -4 to 4
    plt.xticks(x_axis_ticks, x_labels, fontsize=font_size)
    # plt.yticks(y_axis_ticks, y_labels[::-1], fontsize=font_size)
    plt.yticks(y_axis_ticks, y_labels, fontsize=font_size)
    plt.gca().invert_yaxis()

def ant_vis(fig_title):
    # Define the map size
    map_size = 10

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(6, 6))

    # Set the limits of the plot
    ax.set_xlim(0, map_size)
    ax.set_ylim(0, map_size)

    # Plot the red trap at (5,5) with radius 3
    trap_center = (5, 5)
    trap_radius = 2
    trap_circle = plt.Circle(trap_center, trap_radius, color='red', alpha=0.5)
    ax.add_patch(trap_circle)

    # Plot the green goal at (9,9) with radius 1
    goal_center = (8, 8)
    goal_radius = 1
    goal_circle = plt.Circle(goal_center, goal_radius, color='green', alpha=0.5)
    ax.add_patch(goal_circle)

    # Plot the red point at (0,0) with radius 1
    point_center = (2, 2)
    point_radius = 0.5
    point_circle = plt.Circle(point_center, point_radius, color='blue', alpha=0.5)
    ax.add_patch(point_circle)

    # Add the center points for reference
    ax.plot(*trap_center, 'ro')  # Red center point for the trap
    ax.plot(*goal_center, 'go')  # Green center point for the goal
    ax.plot(*point_center, 'b>')  # Red center point for the point

    # Add grid
    ax.grid(True)

    # Add labels
    ax.set_xlabel('X-axis')
    ax.set_ylabel('Y-axis')
    ax.set_title(fig_title)


def swimmer_vis(fig_title):
    # Define the map size
    map_size = 6

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(6, 6))

    # Set the limits of the plot
    ax.set_xlim(0, map_size)
    ax.set_ylim(0, map_size)

    # Plot the red trap at (5,5) with radius 3
    trap_center = (3, 3)
    trap_radius = 1
    trap_circle = plt.Circle(trap_center, trap_radius, color='red', alpha=0.5)
    ax.add_patch(trap_circle)

    # Plot the green goal at (9,9) with radius 1
    goal_center = (5, 5)
    goal_radius = 1
    goal_circle = plt.Circle(goal_center, goal_radius, color='green', alpha=0.5)
    ax.add_patch(goal_circle)

    # Plot the red point at (0,0) with radius 1
    point_center = (1, 1)
    point_radius = 0.5
    point_circle = plt.Circle(point_center, point_radius, color='blue', alpha=0.5)
    ax.add_patch(point_circle)

    # Add the center points for reference
    ax.plot(*trap_center, 'ro')  # Red center point for the trap
    ax.plot(*goal_center, 'go')  # Green center point for the goal
    ax.plot(*point_center, 'b>')  # Red center point for the point

    # Add grid
    ax.grid(True)

    # Add labels
    ax.set_xlabel('X-axis')
    ax.set_ylabel('Y-axis')
    ax.set_title(fig_title)


class EnvSampler():
    def __init__(self, env, max_path_length=1000):
        self.env = env

        self.path_length = 0
        self.current_state = None
        self.max_path_length = max_path_length
        self.path_rewards = []
        self.sum_reward = 0

    def sample(self, agent, eval_t=False, random_explore=False):
        if self.current_state is None:
            if isinstance(self.env.unwrapped, PointGuardedMazeEnv):
                self.current_state = self.env.reset()
                self.current_state = self.current_state[0]['observation']
            else:
                self.current_state = self.env.reset()

        cur_state = self.current_state
        if not random_explore:
            action = agent.select_action(self.current_state, eval_t)
        else:
            action = self.env.action_space.sample()
        if isinstance(self.env.unwrapped, PointGuardedMazeEnv):
            next_state, reward, terminal, _, info = self.env.step(action)
            next_state = next_state['observation']
        else:
            next_state, reward, terminal, info = self.env.step(action)
        # if eval_t:
        #     self.env.render()
        self.path_length += 1
        self.sum_reward += reward

        if terminal or self.path_length >= self.max_path_length:
            self.current_state = None
            self.path_length = 0
            self.path_rewards.append(self.sum_reward)
            self.sum_reward = 0
        else:
            self.current_state = next_state

        return cur_state, action, next_state, reward, terminal, info