import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as data
import torch.optim as optim
import torchvision.models as models
import utils
from tqdm import tqdm
from pathlib import Path
import os
import pickle

LOG_SIG_MAX = 2
LOG_SIG_MIN = -20

def index_batch(batch, indices):
    indexed = {}
    for key in batch.keys():
        indexed[key] = batch[key][indices, ...]
    return indexed


def gen_net(in_size=1, out_size=1, H=128, n_layers=3, activation='tanh'):
    net = []
    for i in range(n_layers):
        net.append(nn.Linear(in_size, H))
        net.append(nn.LeakyReLU())
        in_size = H
    net.append(nn.Linear(in_size, out_size))
    if activation == 'tanh':
        net.append(nn.Tanh())
    elif activation == 'sig':
        net.append(nn.Sigmoid())
    else:
        pass

    return net


def get_winlose_dict(dataset):
    dict = {}
    for i in range(len(dataset['labels'])):
        p1 = dataset['start_indices'][i]
        p2 = dataset['start_indices_2'][i]
        pref = dataset['labels'][i]

        if (p1, p2) not in dict:
            dict[(p1, p2)] = {'p1_wins': 0, 'p2_wins': 0}

        if pref[0] == 1 and pref[1] == 0:
            dict[(p1, p2)]['p1_wins'] += 1
        elif pref[0] == 0 and pref[1] == 1:
            dict[(p1, p2)]['p2_wins'] += 1

    return dict


def get_winlose_times(dict, player1_id, player2_id):
    # Check the record where player1_id is first
    if (player1_id, player2_id) in dict:
        wins = dict[(player1_id, player2_id)]['p1_wins']
        losses = dict[(player1_id, player2_id)]['p2_wins']
    # Check the record where player1_id is second
    elif (player2_id, player1_id) in dict:
        wins = dict[(player2_id, player1_id)]['p2_wins']
        losses = dict[(player2_id, player1_id)]['p1_wins']
    else:
        return (0, 0)  # No games played between these two players

    return wins, losses

# probability alignment models
class MLPRewardModel(object):
    def __init__(self, task, observation_dim, action_dim, ensemble_size=3, lr=3e-4, activation="tanh", logger=None,
                 device="cuda"):
        self.task = task
        self.observation_dim = observation_dim  # state: env.observation_space.shape[0]
        self.action_dim = action_dim  # state: env.action_space.shape[0]
        self.ensemble_size = ensemble_size  # ensemble_size
        self.lr = lr  # learning rate
        self.logger = logger  # logger
        self.device = torch.device(device)

        # build network
        self.opt = None
        self.activation = activation
        self.ensemble = []
        self.paramlst = []
        self.construct_ensemble()

    def construct_ensemble(self):
        for i in range(self.ensemble_size):
            model = nn.Sequential(*gen_net(in_size=self.observation_dim + self.action_dim,
                                           out_size=1, H=256, n_layers=3,
                                           activation=self.activation)).float().to(self.device)
            self.ensemble.append(model)
            self.paramlst.extend(model.parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def save_model(self, path):
        state_dicts = [model.state_dict() for model in self.ensemble]
        torch.save(state_dicts, path)

    def load_model(self, path):
        state_dicts = torch.load(path, map_location='cpu')
        for model, state_dict in zip(self.ensemble, state_dicts):
            model.load_state_dict(state_dict)
            model.to(self.device)

    def train(self, n_epochs, pref_dataset, data_size, batch_size):
        interval = int(data_size / batch_size) + 1

        for epoch in range(1, n_epochs + 1):
            ensemble_losses = [[] for _ in range(self.ensemble_size)]
            ensemble_acc = [[] for _ in range(self.ensemble_size)]

            batch_shuffled_idx = []
            for _ in range(self.ensemble_size):
                batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))

            for i in range(interval):
                self.opt.zero_grad()
                total_loss = 0
                start_pt = i * batch_size
                end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
                for member in range(self.ensemble_size):
                    # get batch
                    batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
                    # compute loss
                    curr_loss, correct = self._train(batch, member)
                    total_loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    ensemble_acc[member].append(correct)
                total_loss.backward()
                self.opt.step()

            train_metrics = {"epoch": epoch,
                             "avg_loss": np.mean(ensemble_losses),
                             "avg_acc": np.mean(ensemble_acc)}
            for i in range(self.ensemble_size):
                train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
                train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
            self.logger.log(train_metrics)

            # early stop
            if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
                break

            # save per 10 epoch
            if (epoch-1) % 10 == 0:
                dir_path = self.logger._model_dir / f'epoch_{epoch-1}'
                try:
                    os.makedirs(dir_path)
                except OSError:
                    pass
                self.save_model(dir_path / f"reward_model.pt")

    def _train(self, batch, member):
        # get batch
        obs_1 = batch['observations']  # batch_size * len_query * obs_dim
        act_1 = batch['actions']  # batch_size * len_query * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
        s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
        s_a_2 = np.concatenate([obs_2, act_2], axis=-1)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get logits
        r_hat1 = self.r_hat_member(s_a_1, member)  # batch_size * len_query * 1
        r_hat2 = self.r_hat_member(s_a_2, member)
        r_hat1 = r_hat1.sum(axis=1)  # batch_size * 1
        r_hat2 = r_hat2.sum(axis=1)
        r_hat = torch.cat([r_hat1, r_hat2], axis=1)  # batch_size * 2

        # get labels
        # labels = torch.from_numpy(labels).long().to(self.device)  # TODO
        labels = torch.from_numpy(labels).to(self.device)

        # compute loss
        curr_loss = self.softXEnt_loss(r_hat, labels)

        # compute acc
        _, predicted = torch.max(r_hat.data, 1)

        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
        return curr_loss, correct

    def r_hat_member(self, x, member):
        return self.ensemble[member](torch.from_numpy(x).float().to(self.device))

    def get_reward_batch(self, x):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized.
        # but I don't understand how the normalization should be happening right now :(
        r_hats = []
        for member in range(self.ensemble_size):
            r_hats.append(self.r_hat_member(x, member=member).detach().cpu().numpy())
        r_hats = np.array(r_hats)

        return np.mean(r_hats, axis=0)

    def softXEnt_loss(self, input, target):
        logprobs = nn.functional.log_softmax(input, dim=1)
        return -(target * logprobs).sum() / input.shape[0]


class CNNRewardModel(MLPRewardModel):
    def __init__(self, task, observation_dim, action_dim, ensemble_size=3, lr=3e-4, activation=None, logger=None,
                 device="cpu"):
        super().__init__(task, observation_dim, action_dim, ensemble_size, lr, activation, logger, device)
        # observation_dim (84, 84)
        # action_dim 6

    def construct_ensemble(self):
        for i in range(self.ensemble_size):
            model = nn.Sequential(
                # CNN Layers
                nn.Conv2d(1, 32, kernel_size=8, stride=4),
                nn.ReLU(),
                nn.Conv2d(32, 64, kernel_size=4, stride=2),
                nn.ReLU(),
                nn.Conv2d(64, 64, kernel_size=3, stride=1),
                nn.ReLU(),
                nn.Flatten(),

                # MLP layers
                *gen_net(in_size=7 * 7 * 64,
                         out_size=1, H=256, n_layers=3,
                         activation=None)
            ).float().to(self.device)

            self.ensemble.append(model)
            self.paramlst.extend(model.parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def r_hat_mlp_member(self, x, member):
        # x should have shape: (batch_size, len_query, 7*7*64 + action_dim)
        return self.ensemble[member][7:](x)

    def r_hat_conv_member(self, observation, member):
        # x should have shape: (batch_size, len_query, 1, 84, 84)
        batch_size = observation.shape[0]
        len_query = observation.shape[1]
        obs_dim = observation.shape[-3:]
        observation = torch.from_numpy(observation).float().to(self.device) / 255.0
        observation = observation.view((-1,) + obs_dim)
        embedding = self.ensemble[member][:7](observation).view(batch_size, len_query, -1)
        return embedding

    def _train(self, batch, member):
        # get batch
        batch_size = batch['observations'].shape[0]
        obs_1 = batch['observations']  # batch_size * len_query * 1 * obs_dim
        act_1 = batch['actions']  # batch_size * len_query * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)

        s_embedding_1 = self.r_hat_conv_member(obs_1, member)  # batch_size * len_query * (7*7*64)
        s_embedding_2 = self.r_hat_conv_member(obs_2, member)  # batch_size * len_query * (7*7*64)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get logits
        r_hat1 = self.r_hat_mlp_member(s_embedding_1, member)  # batch_size * len_query * 1
        r_hat2 = self.r_hat_mlp_member(s_embedding_2, member)
        r_hat1 = r_hat1.sum(axis=1)  # batch_size * 1
        r_hat2 = r_hat2.sum(axis=1)
        r_hat = torch.cat([r_hat1, r_hat2], axis=1)  # batch_size * 2

        # get labels
        # labels = torch.from_numpy(labels).long().to(self.device)  # TODO
        labels = torch.from_numpy(labels).to(self.device)

        # compute loss
        curr_loss = self.softXEnt_loss(r_hat, labels)

        # compute acc
        _, predicted = torch.max(r_hat.data, 1)
        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
        return curr_loss, correct


    def split_train(self, n_epochs, pref_dataset, data_size, batch_size):
        N_DATASET_PARTITION = len(pref_dataset)
        for n_epoch in range(n_epochs):
            ensemble_losses = [[] for _ in range(self.ensemble_size)]
            ensemble_acc = [[] for _ in range(self.ensemble_size)]

            for partition_idx in range(N_DATASET_PARTITION):
                pref_dataset_partition = pref_dataset[partition_idx]
                data_size = pref_dataset_partition["observations"].shape[0]

                interval = int(data_size / batch_size) + 1

                batch_shuffled_idx = []
                for _ in range(self.ensemble_size):
                    batch_shuffled_idx.append(np.random.permutation(pref_dataset_partition["observations"].shape[0]))

                for i in range(interval):
                    self.opt.zero_grad()
                    total_loss = 0
                    start_pt = i * batch_size
                    end_pt = min((i + 1) * batch_size, pref_dataset_partition["observations"].shape[0])
                    for member in range(self.ensemble_size):
                        # get batch
                        batch = index_batch(pref_dataset_partition, batch_shuffled_idx[member][start_pt:end_pt])
                        # compute loss
                        curr_loss, correct = self._train(batch, member)
                        total_loss += curr_loss
                        ensemble_losses[member].append(curr_loss.item())
                        ensemble_acc[member].append(correct)

                    total_loss.backward()
                    self.opt.step()

            train_metrics = {"epoch": n_epoch,
                             "avg_loss": np.mean(ensemble_losses),
                             "avg_acc": np.mean(ensemble_acc)}
            for i in range(self.ensemble_size):
                train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
                train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
            self.logger.log(train_metrics)
    
    def get_reward_batch(self, state):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized.
        # but I don't understand how the normalization should be happening right now :(
        state = np.expand_dims(state, axis=0)
        r_hats = []
        for member in range(self.ensemble_size):
            s_embedding = self.r_hat_conv_member(state, member)
            r_hats.append(self.r_hat_mlp_member(s_embedding, member).detach().cpu().numpy())
        r_hats = np.array(r_hats)

        return np.mean(r_hats, axis=0)


class TransformerRewardModel(MLPRewardModel):
    def __init__(self,
                 task, observation_dim, action_dim, structure_type="transformer1",
                 ensemble_size=3, lr=0.0003, activation="tanh",
                 d_model=256, nhead=4, num_layers=1, max_seq_len=100,
                 logger=None, device="cuda"):
        self.structure_type = structure_type
        self.d_model = d_model
        self.nhead = nhead
        self.num_layers = num_layers
        self.max_seq_len = max_seq_len
        super().__init__(
            task, observation_dim, action_dim,
            ensemble_size, lr, activation, logger, device)
        
    def construct_ensemble(self):
        for i in range(self.ensemble_size):
            if self.structure_type == "transformer1" or self.structure_type == "transformer":
                transformer = utils.PrefTransformer1(
                    self.observation_dim, self.action_dim,
                    self.max_seq_len,
                    self.d_model, self.nhead, self.num_layers
                )
            elif self.structure_type == "transformer2":
                transformer = utils.PrefTransformer2(
                    self.observation_dim, self.action_dim,
                    self.d_model, self.nhead, self.num_layers
                )
            elif self.structure_type == "transformer3":
                transformer = utils.PrefTransformer3(
                    self.observation_dim, self.action_dim,
                    self.d_model, self.nhead, self.num_layers
                )
            else:
                raise NotImplementedError

            self.ensemble.append(transformer.to(self.device))
            self.paramlst.extend(self.ensemble[-1].parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def _train(self, batch, member):
        # get batch
        obs_1 = batch['observations']  # batch_size * len_query * obs_dim
        act_1 = batch['actions']  # batch_size * len_query * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)

        # to_torch
        obs_1 = utils.to_torch(obs_1).to(self.device)
        act_1 = utils.to_torch(act_1).to(self.device)
        obs_2 = utils.to_torch(obs_2).to(self.device)
        act_2 = utils.to_torch(act_2).to(self.device)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get logits
        r_hat1 = self.ensemble[member](obs_1, act_1)  # batch_size * len_query
        r_hat2 = self.ensemble[member](obs_2, act_2)

        # TODO check whether mean or sum?
        # r_hat1 = r_hat1.mean(-1, keepdim=True)  # batch_size * 1
        # r_hat2 = r_hat2.mean(-1, keepdim=True)
        r_hat1 = r_hat1.sum(axis=1).unsqueeze(-1)  # batch_size * 1
        r_hat2 = r_hat2.sum(axis=1).unsqueeze(-1)
        r_hat = torch.cat([r_hat1, r_hat2], axis=-1)  # batch_size * 2
        
        p_1_2 = 1./(1.+torch.exp(r_hat2-r_hat1)) # batch_size * 1
        y = utils.to_torch(labels[:, :1], dtype=torch.float32).to(self.device) # batch_size * 1

        weights = torch.ones_like(y)
        weights[torch.where(y==0.5)] = 0.0
        
        curr_loss = - (weights*(y*torch.log(p_1_2+1e-8) + (1-y)*torch.log(1-p_1_2+1e-8))).mean()
        
        # labels = utils.to_torch(labels, dtype=torch.long).to(self.device)

        # # compute loss
        # curr_loss = self.softXEnt_loss(r_hat, labels)

        # compute acc
        _, predicted = torch.max(r_hat.data, 1)

        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
            
        return curr_loss, correct

# iteration alignment model
class ItrMLPDiffusionRewardModel(MLPRewardModel):
    def __init__(self, task, observation_dim, action_dim, ensemble_size=3, lr=3e-4, activation=None, logger=None,
                 device="cuda",
                 # diffusion parameters
                 num_step = 200, beta_start = 1e-4, beta_end = 1e-1, traj_len = 200, data_dim = 1):
        self.num_step = num_step
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.traj_len = traj_len
        self.data_dim = data_dim
        super().__init__(task, observation_dim, action_dim, ensemble_size, lr, activation, logger, device)

    def construct_ensemble(self):
        for i in range(self.ensemble_size):

            diffusion = utils.MLPDiffusion(
                self.observation_dim, self.action_dim,
                self.num_step,
                self.beta_start,
                self.beta_end,
                self.traj_len,
                self.data_dim,
                self.device,
            )

            self.ensemble.append(diffusion.to(self.device))
            self.paramlst.extend(self.ensemble[-1].parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def _train(self, pref_dataset, batch, member, first):
        # TODO
        # now w_ij and w_ji = 0 or 1 in each pair, need to calculate in prior do in the future
        # and need to handle tie (delete)

        # get batch
        obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
        act_1 = batch['actions']  # batch_size * traj_len * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
        # s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
        # s_a_2 = np.concatenate([obs_2, act_2], axis=-1)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get logits
        # if first == True:
        #     r_hat1 = torch.ones(obs_1.shape[0], obs_1.shape[1], 1, device='cuda')
        #     r_hat2 = torch.ones(obs_1.shape[0], obs_1.shape[1], 1, device='cuda')
        # else:
        r_hat1 = self.r_hat_member(obs_1, act_1, member)  # batch_size * traj_len * 1
        r_hat2 = self.r_hat_member(obs_2, act_2, member)

        r_hat1_sum = r_hat1.sum(axis=1)  # batch_size * 1
        r_hat2_sum = r_hat2.sum(axis=1)

        # should calculate before normalization for correct comparison
        r_hat_sum = torch.cat([r_hat1_sum, r_hat2_sum], axis=1)  # batch_size * 2

        # TODO
        # normalize to satisfy geometric mean strength = 1
        r_hat1 = (r_hat1 - r_hat1.mean()) / (r_hat1.std() + 1e-8)
        r_hat2 = (r_hat2 - r_hat2.mean()) / (r_hat2.std() + 1e-8)
        r_hat1_sum = (r_hat1_sum - r_hat1_sum.mean()) / (r_hat1_sum.std() + 1e-8)
        r_hat2_sum = (r_hat2_sum - r_hat2_sum.mean()) / (r_hat2_sum.std() + 1e-8)

        # calculate iterative objective
        target_traj_reward = torch.zeros_like(r_hat1_sum)  # batch_size * 1
        target_step_reward = torch.zeros_like(r_hat1) # batch_size * traj_len * 1

        for i in range(len(batch['observations'])):

            sum_1 = 0
            sum_2 = 0

            reward_i = r_hat1_sum[i]
            index_i = batch['start_indices'][i]

            for j in range(len(batch['observations_2'])):
                reward_j = r_hat2_sum[j]
                index_j = batch['start_indices_2'][j]
                # need to search in the whole dataset, not only the batch
                for k in range(len(pref_dataset['observations'])):
                    if index_i == pref_dataset['start_indices'][k] and index_j == pref_dataset['start_indices_2'][k]:
                        if pref_dataset['labels'][k][0] > pref_dataset['labels'][k][1]:
                            sum_1 += torch.exp(reward_j) / (torch.exp(reward_i) + torch.exp(reward_j))
                        elif pref_dataset['labels'][k][0] < pref_dataset['labels'][k][1]:
                            sum_2 += 1 / (torch.exp(reward_i) + torch.exp(reward_j))

            target_traj_reward[i] = torch.log(((1/(torch.exp(reward_i)+1)) + sum_1) / ((1/(torch.exp(reward_i)+1)) + sum_2))

            # for j in range(len(batch['observations_2'])):
            #     reward_j = r_hat2_sum[j]
            #     index_j = batch['start_indices_2'][j]
            #     # need to search in the whole dataset, not only the batch
            #     for k in range(len(pref_dataset['observations'])):
            #         if index_i == pref_dataset['start_indices'][k] and index_j == pref_dataset['start_indices_2'][k]:
            #             if pref_dataset['labels'][k][0] > pref_dataset['labels'][k][1]:
            #                 sum_1 += reward_j / (reward_i + reward_j)
            #             elif pref_dataset['labels'][k][0] < pref_dataset['labels'][k][1]:
            #                 sum_2 += 1 / (reward_i + reward_j)
            #
            # target_traj_reward[i] = ((1/(reward_i+1)) + sum_1) / ((1/(reward_i+1)) + sum_2)

            # print('---------------------------------')
            # print(r_hat1[[i]])
            # print(reward_i)
            # print((torch.exp(reward_i)+1))
            # print(sum_1)
            # print((torch.exp(reward_i)+1))
            # print(sum_2)
            # print('---------------------------------')
            delta = target_traj_reward[i] - reward_i
            average_delta = delta / self.traj_len
            target_step_reward[i] = r_hat1[i] + average_delta


        curr_loss = self.ensemble[member].compute_loss(
            obs_1,
            act_1,
            x_0=target_step_reward.to(self.device),
        )
        print(curr_loss)

        # compute acc
        _, predicted = torch.max(r_hat_sum.data, 1)

        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
        return curr_loss, correct

    def train(self, n_epochs, pref_dataset, data_size, batch_size):
        interval = int(data_size / batch_size) + 1

        for epoch in tqdm(range(1, n_epochs + 1)):
            ensemble_losses = [[] for _ in range(self.ensemble_size)]
            ensemble_acc = [[] for _ in range(self.ensemble_size)]

            batch_shuffled_idx = []
            for _ in range(self.ensemble_size):
                batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))

            for i in tqdm(range(interval)):
                self.opt.zero_grad()
                total_loss = 0
                start_pt = i * batch_size
                end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
                for member in range(self.ensemble_size):
                    # get batch
                    batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
                    # compute loss
                    if epoch == 1 and i == 0:
                        first = True
                    else:
                        first = False
                    curr_loss, correct = self._train(pref_dataset, batch, member, first)
                    total_loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    ensemble_acc[member].append(correct)
                total_loss.backward()
                self.opt.step()

            train_metrics = {"epoch": epoch,
                             "avg_loss": np.mean(ensemble_losses),
                             "avg_acc": np.mean(ensemble_acc)}
            for i in range(self.ensemble_size):
                train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
                train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
            self.logger.log(train_metrics)

            # early stop
            if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
                break

    def r_hat_member(self, obs, acs, member):
        obs = torch.from_numpy(obs).float().to(self.device)
        acs = torch.from_numpy(acs).float().to(self.device)
        reward = self.ensemble[member].sample(obs, acs, during_training=False)
        return reward

    def get_reward_batch(self, x):

        print('need to write')
        exit()

# # this version search in the whole preference dataset
# class ItrMLPRewardModel(MLPRewardModel):
#     def __init__(self, task, observation_dim, action_dim, ensemble_size=3, lr=3e-4, activation="tanh", logger=None,
#                  device="cuda", **kwargs):
#         self.reward_model_type = kwargs.get('reward_model_type', 'DistributionalMLP')
#         super().__init__(task, observation_dim, action_dim, ensemble_size, lr, activation, logger, device)
#         self.task = task
#         self.observation_dim = observation_dim  # state: env.observation_space.shape[0]
#         self.action_dim = action_dim  # state: env.action_space.shape[0]
#         self.ensemble_size = ensemble_size  # ensemble_size
#         self.lr = lr  # learning rate
#         self.logger = logger  # logger
#         self.device = torch.device(device)
#
#         self.risk_sensitive = kwargs.get('risk_sensitive', False)
#         self.beta_model_path = kwargs.get('beta_model_path', None)
#         self.beta_lr = kwargs.get('beta_lr', 3e-5)
#         self.beta_n_epochs = kwargs.get('beta_n_epochs', 200)
#         self.beta_model_type = kwargs.get('beta_model_type', 'MLP')
#         self.beta_coef = kwargs.get('beta_coeff', 0.01)
#         self.beta_discrete = kwargs.get('beta_discrete', False)
#
#         # build network
#         self.opt = None
#         self.activation = activation
#         self.ensemble = []
#         self.paramlst = []
#         self.construct_ensemble()
#
#     def construct_ensemble(self):
#         if 'Distributional' in self.reward_model_type:
#             for i in range(self.ensemble_size):
#                 model = nn.Sequential(*gen_net(in_size=self.observation_dim + self.action_dim,
#                                                out_size=2, H=256, n_layers=3,
#                                                activation=self.activation)).float().to(self.device)
#                 self.ensemble.append(model)
#                 self.paramlst.extend(model.parameters())
#         else:
#             for i in range(self.ensemble_size):
#                 model = nn.Sequential(*gen_net(in_size=self.observation_dim + self.action_dim,
#                                                out_size=1, H=256, n_layers=3,
#                                                activation=self.activation)).float().to(self.device)
#                 self.ensemble.append(model)
#                 self.paramlst.extend(model.parameters())
#
#         self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)
#
#     def train(self, n_epochs, pref_dataset, data_size, batch_size):
#         # preprocess pref_dataset to get win_lose_dict
#         win_lose_dict = get_winlose_dict(pref_dataset)
#
#         # train beta model
#         if self.risk_sensitive:
#             self.beta_model = utils.BetaNetwork(self.observation_dim, self.action_dim,
#                                                 self.beta_lr, self.device, self.beta_model_type, self.beta_coef)
#             # continuous -- beta transformer
#             if not self.beta_discrete:
#                 save_dir = Path(self.beta_model_path) / f'{self.task}' / f'{self.beta_model_type}_Coef{self.beta_coef}_lr{self.beta_lr}'
#                 model_dir = save_dir / f'itr_{self.beta_n_epochs}'
#                 model_dir.mkdir(parents=True, exist_ok=True)
#                 model_file = model_dir / 'beta_model.pth'
#                 if not model_file.is_file():
#                     # train beta model
#                     print('Start training Beta model.')
#                     self.beta_model.fit_data(num_epochs=self.beta_n_epochs, data_size=data_size,
#                                              batch_size=batch_size, dataset=pref_dataset, save_dir=save_dir)
#                     print('Finish training Beta model.')
#                     print(f'Save Beta model in {save_dir}.')
#                     print('PLease rerun the code to load the trained Beta model.')
#                     exit()
#                     # self.beta_model.save_model(model_file)
#                 else:
#                     # load beta model
#                     print(f'Load Beta model from {model_file}.')
#                     self.beta_model.load_model(model_file)
#             # discrete -- beta dict
#             else:
#                 print('Start learning Beta dict.')
#                 self.beta_model.fit_data_discrete(dataset=pref_dataset)
#                 print('Finish learning Beta dict.')
#
#         # train reward model
#         print('Start training reward model.')
#         interval = int(data_size / batch_size) + 1
#
#         for epoch in range(1, n_epochs + 1):
#             ensemble_losses = [[] for _ in range(self.ensemble_size)]
#             ensemble_acc = [[] for _ in range(self.ensemble_size)]
#
#             batch_shuffled_idx = []
#             for _ in range(self.ensemble_size):
#                 batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))
#
#             for i in tqdm(range(interval)):
#                 self.opt.zero_grad()
#                 total_loss = 0
#                 start_pt = i * batch_size
#                 end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
#                 for member in range(self.ensemble_size):
#                     # get batch
#                     batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
#                     # compute loss
#                     curr_loss, correct = self._train(pref_dataset, win_lose_dict, batch, member)
#                     total_loss += curr_loss
#                     ensemble_losses[member].append(curr_loss.item())
#                     ensemble_acc[member].append(correct)
#                 total_loss.backward()
#                 self.opt.step()
#
#             train_metrics = {"epoch": epoch,
#                              "avg_loss": np.mean(ensemble_losses),
#                              "avg_acc": np.mean(ensemble_acc)}
#             for i in range(self.ensemble_size):
#                 train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
#                 train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
#             self.logger.log(train_metrics)
#
#             # early stop
#             if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
#                 break
#
#             # save per 10 epoch
#             if (epoch-1) % 10 == 0:
#                 dir_path = self.logger._model_dir / f'epoch_{epoch-1}'
#                 try:
#                     os.makedirs(dir_path)
#                 except OSError:
#                     pass
#                 self.save_model(dir_path / f"reward_model.pt")
#
#         print('Finish training reward model.')
#
#     def _train(self, pref_dataset, win_lose_dict, batch, member):
#         # TODO
#         # maybe need to handle with tie in the future (now I just skip these pairs)
#
#         # get batch
#         obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
#         act_1 = batch['actions']  # batch_size * traj_len * action_dim
#         obs_2 = batch['observations_2']
#         act_2 = batch['actions_2']
#         labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
#         s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
#         s_a_2 = np.concatenate([obs_2, act_2], axis=-1)
#
#         # get comparable labels
#         comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
#         comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)
#
#         # get step-wise rewards
#         r_hat1 = self.r_hat_member(s_a_1, member)  # batch_size * traj_len * 1
#         r_hat2 = self.r_hat_member(s_a_2, member)
#
#         r_hat1_sum = r_hat1.sum(axis=1)  # batch_size * 1
#         r_hat2_sum = r_hat2.sum(axis=1)
#
#         # should calculate before normalization for correct comparison
#         # r_hat = torch.cat([r_hat1, r_hat2], axis=2)
#         r_hat_sum = torch.cat([r_hat1_sum, r_hat2_sum], axis=1)  # batch_size * 2
#
#         # TODO
#         # normalize to satisfy geometric mean strength = 1
#         # r_hat1 = (r_hat1 - r_hat.mean()) / (r_hat.std() + 1e-8)
#         # r_hat2 = (r_hat2 - r_hat.mean()) / (r_hat.std() + 1e-8)
#         r_hat1_sum = (r_hat1_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#         # r_hat2_sum = (r_hat2_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#
#
#         # calculate iterative objective
#         target_traj_reward = torch.zeros_like(r_hat1_sum)  # batch_size * 1
#         # target_step_reward = torch.zeros_like(r_hat1) # batch_size * traj_len * 1
#
#         # note that 'index' (i/j) denotes the index in the batch
#         # while 'start_index' denotes the starting timestep in the pref_data
#         unique_i_start_indices = np.unique(batch['start_indices'])
#         for start_index_i in unique_i_start_indices:
#
#             sum_1 = 0
#             sum_2 = 0
#
#             i = np.where(batch['start_indices'] == start_index_i)[0]
#
#             # avoid that the batch contains a trajectory many times (many identical i in the batch)
#             # they have the same start_indices but different batch index, calculate once is enough
#             if len(i) > 1:
#                 i = np.array([i[0]])
#
#             reward_i = r_hat1_sum[i].squeeze(-1)
#             exp_reward_i = torch.exp(reward_i)
#
#             start_index_j_list = []
#
#             # given an index i, only search j in the existing pair_labels (i,j,label) or (j,i,label)
#             for pair in win_lose_dict.keys():
#                 # order (i,j) '0'
#                 if pair[0] == start_index_i:
#                     start_index_j_list.append((pair[1], 0))
#                 # order (j,i) '1'
#                 elif pair[1] == start_index_i:
#                     start_index_j_list.append((pair[0], 1))
#
#             # automatically remove duplicates
#             unique_j_start_indices = set(start_index_j_list)
#
#             for (start_index_j, order) in unique_j_start_indices:
#                 obs_key = 'observations_2' if order == 0 else 'observations'
#                 act_key = 'actions_2' if order == 0 else 'actions'
#                 start_indices_key = 'start_indices_2' if order == 0 else 'start_indices'
#
#                 j = np.where(pref_dataset[start_indices_key] == start_index_j)[0]
#                 # avoid that the dataset contains a trajectory many times (many identical j in the batch)
#                 # they have the same start_indices but different batch index, calculate once is enough
#                 if len(j) > 1:
#                     j = np.array([j[0]])
#
#                 s_a_j = np.concatenate([pref_dataset[obs_key][j], pref_dataset[act_key][j]], axis=-1)
#
#                 reward_j = self.r_hat_member(s_a_j, member).sum(axis=1).squeeze(-1)
#                 reward_j = (reward_j - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#                 exp_reward_j = torch.exp(reward_j)
#
#                 w_ij, w_ji = get_winlose_times(win_lose_dict, start_index_i, start_index_j)
#
#                 sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
#                 sum_2 += w_ji / (exp_reward_i + exp_reward_j)
#             if self.risk_sensitive:
#                 if not self.beta_discrete:
#                     alpha, beta = self.beta_model.get_alpha_beta(torch.from_numpy(s_a_1[i]).float().to(self.device))
#                 else:
#                     alpha, beta = self.beta_model.get_rescaled_alpha_beta_discrete\
#                         (torch.from_numpy(s_a_1[i]).float().to(self.device), rescale=10)
#                 #print(f'alpha_{alpha.item()}_beta_{beta.item()}')
#
#                 target_traj_reward[i] = torch.log(((alpha/(exp_reward_i+1)) + sum_1) + 1e-8) - \
#                                         torch.log(((beta/(exp_reward_i+1)) + sum_2) + 1e-8)
#             else:
#                 target_traj_reward[i] = torch.log(((1/(exp_reward_i+1)) + sum_1) + 1e-8) - \
#                                         torch.log(((1/(exp_reward_i+1)) + sum_2) + 1e-8)
#
#         curr_loss = F.mse_loss(r_hat1_sum, target_traj_reward)
#         # print(curr_loss)
#
#
#         # compute acc
#         _, predicted = torch.max(r_hat_sum.data, 1)
#
#         if not len(comparable_indices):
#             correct = 0.7  # TODO, for exception
#         else:
#             correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
#                 comparable_indices)
#         return curr_loss, correct
#
#     def r_hat_member(self, x, member):
#         if 'Distributional' in self.reward_model_type:
#             x = self.ensemble[member](torch.from_numpy(x).float().to(self.device))
#             mean = x[:,:,0]
#             log_std = x[:,:,1]
#             log_std = torch.clamp(log_std, min=LOG_SIG_MIN, max=LOG_SIG_MAX)
#             std = log_std.exp()
#             normal = torch.distributions.Normal(mean, std)
#             reward = normal.rsample().unsqueeze(-1)
#             return reward
#         else:
#             return self.ensemble[member](torch.from_numpy(x).float().to(self.device))
#
#     def get_reward_batch(self, x):
#         # they say they average the rewards from each member of the ensemble,
#         # but I think this only makes sense if the rewards are already normalized.
#         # but I don't understand how the normalization should be happening right now :(
#         r_hats = []
#         if 'Distributional' in self.reward_model_type:
#             for member in range(self.ensemble_size):
#                 r_hats.append(self.r_hat_member_dis(x, member=member).detach().cpu().numpy())
#         else:
#             for member in range(self.ensemble_size):
#                 r_hats.append(self.r_hat_member(x, member=member).detach().cpu().numpy())
#         r_hats = np.array(r_hats)
#
#         return np.mean(r_hats, axis=0)
#
#     def r_hat_member_dis(self, x, member):
#         x = self.ensemble[member](torch.from_numpy(x).float().to(self.device))
#         mean = x[:,0]
#         log_std = x[:,1]
#         log_std = torch.clamp(log_std, min=LOG_SIG_MIN, max=LOG_SIG_MAX)
#         std = log_std.exp()
#         normal = torch.distributions.Normal(mean, std)
#         reward = normal.rsample()
#         return reward

# this version only search in the batch
class ItrMLPRewardModel(MLPRewardModel):
    def __init__(self, task, observation_dim, action_dim, ensemble_size=3, lr=3e-4, activation="tanh", logger=None,
                 device="cuda", **kwargs):
        self.reward_model_type = kwargs.get('reward_model_type', 'MLP')
        super().__init__(task, observation_dim, action_dim, ensemble_size, lr, activation, logger, device)
        self.task = task
        self.observation_dim = observation_dim  # state: env.observation_space.shape[0]
        self.action_dim = action_dim  # state: env.action_space.shape[0]
        self.ensemble_size = ensemble_size  # ensemble_size
        self.lr = lr  # learning rate
        self.logger = logger  # logger
        self.device = torch.device(device)

        self.risk_sensitive = kwargs.get('risk_sensitive', False)
        self.beta_model_path = kwargs.get('beta_model_path', None)
        self.beta_lr = kwargs.get('beta_lr', 3e-5)
        self.beta_n_epochs = kwargs.get('beta_n_epochs', 200)
        self.beta_model_type = kwargs.get('beta_model_type', 'MLP')
        self.beta_coef = kwargs.get('beta_coeff', 0.01)
        self.beta_discrete = kwargs.get('beta_discrete', False)

        # build network
        self.opt = None
        self.activation = activation
        self.ensemble = []
        self.paramlst = []
        self.construct_ensemble()

    def construct_ensemble(self):
        if 'Distributional' in self.reward_model_type:
            for i in range(self.ensemble_size):
                model = nn.Sequential(*gen_net(in_size=self.observation_dim + self.action_dim,
                                               out_size=2, H=256, n_layers=3,
                                               activation=self.activation)).float().to(self.device)
                self.ensemble.append(model)
                self.paramlst.extend(model.parameters())
        else:
            for i in range(self.ensemble_size):
                model = nn.Sequential(*gen_net(in_size=self.observation_dim + self.action_dim,
                                               out_size=1, H=256, n_layers=3,
                                               activation=self.activation)).float().to(self.device)
                self.ensemble.append(model)
                self.paramlst.extend(model.parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def train(self, n_epochs, pref_dataset, data_size, batch_size):

        # train beta model
        if self.risk_sensitive:
            self.beta_model = utils.BetaNetwork(self.observation_dim, self.action_dim,
                                                self.beta_lr, self.device, self.beta_model_type, self.beta_coef)
            # continuous -- beta transformer
            if not self.beta_discrete:
                save_dir = Path(self.beta_model_path) / f'{self.task}' / f'{self.beta_model_type}_Coef{self.beta_coef}_lr{self.beta_lr}'
                model_dir = save_dir / f'itr_{self.beta_n_epochs}'
                model_dir.mkdir(parents=True, exist_ok=True)
                model_file = model_dir / 'beta_model.pth'
                if not model_file.is_file():
                    # train beta model
                    print('Start training Beta model.')
                    self.beta_model.fit_data(num_epochs=self.beta_n_epochs, data_size=data_size,
                                             batch_size=batch_size, dataset=pref_dataset, save_dir=save_dir)
                    print('Finish training Beta model.')
                    print(f'Save Beta model in {save_dir}.')
                    print('PLease rerun the code to load the trained Beta model.')
                    exit()
                    # self.beta_model.save_model(model_file)
                else:
                    # load beta model
                    print(f'Load Beta model from {model_file}.')
                    self.beta_model.load_model(model_file)
            # discrete -- beta dict
            else:
                print('Start learning Beta dict.')
                self.beta_model.fit_data_discrete(dataset=pref_dataset)
                print('Finish learning Beta dict.')

        # train reward model
        print('Start training reward model.')
        interval = int(data_size / batch_size) + 1

        for epoch in range(1, n_epochs + 1):
            ensemble_losses = [[] for _ in range(self.ensemble_size)]
            ensemble_acc = [[] for _ in range(self.ensemble_size)]

            batch_shuffled_idx = []
            for _ in range(self.ensemble_size):
                batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))

            for i in tqdm(range(interval)):
                self.opt.zero_grad()
                total_loss = 0
                start_pt = i * batch_size
                end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
                for member in range(self.ensemble_size):
                    # get batch
                    batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
                    # compute loss
                    win_lose_dict = get_winlose_dict(batch)
                    curr_loss, correct = self._train(win_lose_dict, batch, member)
                    total_loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    ensemble_acc[member].append(correct)
                total_loss.backward()
                self.opt.step()

            train_metrics = {"epoch": epoch,
                             "avg_loss": np.mean(ensemble_losses),
                             "avg_acc": np.mean(ensemble_acc)}
            for i in range(self.ensemble_size):
                train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
                train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
            self.logger.log(train_metrics)

            # early stop
            if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
                break

            # save per 10 epoch
            if (epoch-1) % 10 == 0:
                dir_path = self.logger._model_dir / f'epoch_{epoch-1}'
                try:
                    os.makedirs(dir_path)
                except OSError:
                    pass
                self.save_model(dir_path / f"reward_model.pt")

        print('Finish training reward model.')

    def _train(self, win_lose_dict, batch, member):
        # TODO
        # maybe need to handle with tie in the future (now I just skip these pairs)

        # get batch
        obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
        act_1 = batch['actions']  # batch_size * traj_len * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
        s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
        s_a_2 = np.concatenate([obs_2, act_2], axis=-1)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get step-wise rewards
        r_hat1 = self.r_hat_member(s_a_1, member)  # batch_size * traj_len * 1
        r_hat2 = self.r_hat_member(s_a_2, member)

        r_hat1_sum = r_hat1.sum(axis=1)  # batch_size * 1
        r_hat2_sum = r_hat2.sum(axis=1)

        # should calculate before normalization for correct comparison
        # r_hat = torch.cat([r_hat1, r_hat2], axis=2)
        r_hat_sum = torch.cat([r_hat1_sum, r_hat2_sum], axis=1)  # batch_size * 2

        # TODO do not normalize, only force mean
        # normalize to satisfy geometric mean strength = 1
        # r_hat1 = (r_hat1 - r_hat.mean()) / (r_hat.std() + 1e-8)
        # r_hat2 = (r_hat2 - r_hat.mean()) / (r_hat.std() + 1e-8)
        # r_hat1_sum = (r_hat1_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
        # r_hat2_sum = (r_hat2_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
        r_hat1_sum = (r_hat1_sum - r_hat_sum.mean())
        r_hat2_sum = (r_hat2_sum - r_hat_sum.mean())


        # calculate iterative objective
        target_traj_reward = torch.zeros_like(r_hat1_sum)  # batch_size * 1
        # target_step_reward = torch.zeros_like(r_hat1) # batch_size * traj_len * 1

        # note that 'index' (i/j) denotes the index in the batch
        # while 'start_index' denotes the starting timestep in the pref_data
        unique_i_start_indices = np.unique(batch['start_indices'])
        for start_index_i in unique_i_start_indices:

            sum_1 = 0
            sum_2 = 0

            i = np.where(batch['start_indices'] == start_index_i)[0]

            # avoid that the batch contains a trajectory many times (many identical i in the batch)
            # they have the same start_indices but different batch index, calculate once is enough
            if len(i) > 1:
                i = np.array([i[0]])

            reward_i = r_hat1_sum[i].squeeze(-1)
            exp_reward_i = torch.exp(reward_i)

            start_index_j_list = []

            # given an index i, only search j in the existing pair_labels (i,j,label) or (j,i,label)
            for pair in win_lose_dict.keys():
                # order (i,j) '0'
                if pair[0] == start_index_i:
                    start_index_j_list.append((pair[1], 0))
                # order (j,i) '1'
                elif pair[1] == start_index_i:
                    start_index_j_list.append((pair[0], 1))

            # automatically remove duplicates
            unique_j_start_indices = set(start_index_j_list)

            for (start_index_j, order) in unique_j_start_indices:
                start_indices_key = 'start_indices_2' if order == 0 else 'start_indices'

                j = np.where(batch[start_indices_key] == start_index_j)[0]
                # avoid that the dataset contains a trajectory many times (many identical j in the batch)
                # they have the same start_indices but different batch index, calculate once is enough
                if len(j) > 1:
                    j = np.array([j[0]])

                reward_j = r_hat2_sum[j] if order == 0 else r_hat1_sum[j]
                exp_reward_j = torch.exp(reward_j)

                w_ij, w_ji = get_winlose_times(win_lose_dict, start_index_i, start_index_j)

            if self.risk_sensitive:
                if not self.beta_discrete:
                    alpha, beta = self.beta_model.get_alpha_beta(torch.from_numpy(s_a_1[i]).float().to(self.device))
                else:
                    alpha, beta = self.beta_model.get_rescaled_alpha_beta_discrete\
                        (torch.from_numpy(s_a_1[i]).float().to(self.device), rescale=10)
                #print(f'alpha_{alpha.item()}_beta_{beta.item()}')
                with torch.no_grad():
                    for k in range(1):
                        sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
                        sum_2 += w_ji / (exp_reward_i + exp_reward_j)
                        exp_reward_i_next = torch.log(torch.clamp(((alpha /(exp_reward_i+1)) + sum_1), min=1e-8)) - \
                                            torch.log(torch.clamp(((beta /(exp_reward_i+1)) + sum_2), min=1e-8))
                        delta = abs(exp_reward_i_next - exp_reward_i)
                        exp_reward_i = exp_reward_i_next
                        if delta < 1e-3:
                            break
                    target_traj_reward[i] = exp_reward_i
            else:
                with torch.no_grad():
                    for k in range(1):
                        sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
                        sum_2 += w_ji / (exp_reward_i + exp_reward_j)
                        torch.clamp(((1/(exp_reward_i+1)) + sum_1), min=1e-8)
                        exp_reward_i_next = torch.log(torch.clamp(((1/(exp_reward_i+1)) + sum_1), min=1e-8)) - \
                                            torch.log(torch.clamp(((1/(exp_reward_i+1)) + sum_2), min=1e-8))
                        delta = abs(exp_reward_i_next - exp_reward_i)
                        exp_reward_i = exp_reward_i_next
                        if delta < 1e-3:
                            break
                    target_traj_reward[i] = exp_reward_i

        curr_loss = F.mse_loss(r_hat1_sum, target_traj_reward)
        # print(curr_loss)


        # compute acc
        _, predicted = torch.max(r_hat_sum.data, 1)

        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
        return curr_loss, correct

    def r_hat_member(self, x, member):
        if 'Distributional' in self.reward_model_type:
            x = self.ensemble[member](torch.from_numpy(x).float().to(self.device))
            mean = x[:,:,0]
            log_std = x[:,:,1]
            log_std = torch.clamp(log_std, min=LOG_SIG_MIN, max=LOG_SIG_MAX)
            std = log_std.exp()
            normal = torch.distributions.Normal(mean, std)
            reward = normal.rsample().unsqueeze(-1)
            return reward
        else:
            return self.ensemble[member](torch.from_numpy(x).float().to(self.device))

    def get_reward_batch(self, x):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized.
        # but I don't understand how the normalization should be happening right now :(
        r_hats = []
        if 'Distributional' in self.reward_model_type:
            for member in range(self.ensemble_size):
                r_hats.append(self.r_hat_member_dis(x, member=member).detach().cpu().numpy())
        else:
            for member in range(self.ensemble_size):
                r_hats.append(self.r_hat_member(x, member=member).detach().cpu().numpy())
        r_hats = np.array(r_hats)

        return np.mean(r_hats, axis=0)

    def r_hat_member_dis(self, x, member):
        x = self.ensemble[member](torch.from_numpy(x).float().to(self.device))
        mean = x[:,0]
        log_std = x[:,1]
        log_std = torch.clamp(log_std, min=LOG_SIG_MIN, max=LOG_SIG_MAX)
        std = log_std.exp()
        normal = torch.distributions.Normal(mean, std)
        reward = normal.rsample()
        return reward

# # search in the whole pref dataset
# class ItrTransformerRewardModel(MLPRewardModel):
#     def __init__(self,
#                  task, observation_dim, action_dim, structure_type="transformer",
#                  ensemble_size=3, lr=0.0003, activation="tanh",
#                  d_model=256, nhead=4, num_layers=1, max_seq_len=100,
#                  logger=None, device="cuda", **kwargs):
#         self.structure_type = structure_type
#         self.d_model = d_model
#         self.nhead = nhead
#         self.num_layers = num_layers
#         self.max_seq_len = max_seq_len
#         self.reward_model_type = kwargs.get('reward_model_type', 'Transformer')
#
#         super().__init__(
#             task, observation_dim, action_dim,
#             ensemble_size, lr, activation, logger, device)
#
#         self.risk_sensitive = kwargs.get('risk_sensitive', False)
#         self.beta_model_path = kwargs.get('beta_model_path', None)
#         self.beta_lr = kwargs.get('beta_lr', 3e-5)
#         self.beta_n_epochs = kwargs.get('beta_n_epochs', 200)
#         self.beta_model_type = kwargs.get('beta_model_type', 'MLP')
#         self.beta_coef = kwargs.get('beta_coeff', 0.01)
#         self.beta_discrete = kwargs.get('beta_discrete', False)
#
#     def construct_ensemble(self):
#         for i in range(self.ensemble_size):
#             if 'Distributional' in self.reward_model_type:
#                 transformer = utils.DistributionalPrefTransformer(
#                     self.observation_dim, self.action_dim,
#                     self.max_seq_len,
#                     self.d_model, self.nhead, self.num_layers
#                 )
#             else:
#                 transformer = utils.PrefTransformer1(
#                     self.observation_dim, self.action_dim,
#                     self.max_seq_len,
#                     self.d_model, self.nhead, self.num_layers
#                 )
#             self.ensemble.append(transformer.to(self.device))
#             self.paramlst.extend(self.ensemble[-1].parameters())
#
#         self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)
#
#     def train(self, n_epochs, pref_dataset, data_size, batch_size):
#         # preprocess pref_dataset to get win_lose_dict
#         win_lose_dict = get_winlose_dict(pref_dataset)
#
#         # train beta model
#         if self.risk_sensitive:
#             self.beta_model = utils.BetaNetwork(self.observation_dim, self.action_dim,
#                                                 self.beta_lr, self.device, self.beta_model_type, self.beta_coef)
#             # continuous -- beta transformer
#             if not self.beta_discrete:
#                 save_dir = Path(self.beta_model_path) / f'{self.task}' / f'{self.beta_model_type}_Coef{self.beta_coef}_lr{self.beta_lr}'
#                 model_dir = save_dir / f'itr_{self.beta_n_epochs}'
#                 model_dir.mkdir(parents=True, exist_ok=True)
#                 model_file = model_dir / 'beta_model.pth'
#                 if not model_file.is_file():
#                     # train beta model
#                     print('Start training Beta model.')
#                     self.beta_model.fit_data(num_epochs=self.beta_n_epochs, data_size=data_size,
#                                              batch_size=batch_size, dataset=pref_dataset, save_dir=save_dir)
#                     print('Finish training Beta model.')
#                     print(f'Save Beta model in {save_dir}.')
#                     print('PLease rerun the code to load the trained Beta model.')
#                     exit()
#                     # self.beta_model.save_model(model_file)
#                 else:
#                     # load beta model
#                     print(f'Load Beta model from {model_file}.')
#                     self.beta_model.load_model(model_file)
#             # discrete -- beta dict
#             else:
#                 print('Start learning Beta dict.')
#                 self.beta_model.fit_data_discrete(dataset=pref_dataset)
#                 print('Finish learning Beta dict.')
#
#         # train reward model
#         print('Start training reward model.')
#         interval = int(data_size / batch_size) + 1
#
#         for epoch in range(1, n_epochs + 1):
#             ensemble_losses = [[] for _ in range(self.ensemble_size)]
#             ensemble_acc = [[] for _ in range(self.ensemble_size)]
#
#             batch_shuffled_idx = []
#             for _ in range(self.ensemble_size):
#                 batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))
#
#             for i in tqdm(range(interval)):
#                 self.opt.zero_grad()
#                 total_loss = 0
#                 start_pt = i * batch_size
#                 end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
#                 for member in range(self.ensemble_size):
#                     # get batch
#                     batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
#                     # compute loss
#                     curr_loss, correct = self._train(pref_dataset, win_lose_dict, batch, member)
#                     total_loss += curr_loss
#                     ensemble_losses[member].append(curr_loss.item())
#                     ensemble_acc[member].append(correct)
#                 total_loss.backward()
#                 self.opt.step()
#
#             train_metrics = {"epoch": epoch,
#                              "avg_loss": np.mean(ensemble_losses),
#                              "avg_acc": np.mean(ensemble_acc)}
#             for i in range(self.ensemble_size):
#                 train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
#                 train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
#             self.logger.log(train_metrics)
#
#             # early stop
#             if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
#                 break
#
#             # save per 10 epoch
#             if (epoch-1) % 10 == 0:
#                 dir_path = self.logger._model_dir / f'epoch_{epoch-1}'
#                 try:
#                     os.makedirs(dir_path)
#                 except OSError:
#                     pass
#                 self.save_model(dir_path / f"reward_model.pt")
#
#         print('Finish training reward model.')
#
#     def _train(self, pref_dataset, win_lose_dict, batch, member):
#         # TODO
#         # maybe need to handle with tie in the future (now I just skip these pairs)
#
#         # get batch
#         obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
#         act_1 = batch['actions']  # batch_size * traj_len * action_dim
#         obs_2 = batch['observations_2']
#         act_2 = batch['actions_2']
#         labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
#         s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
#         s_a_2 = np.concatenate([obs_2, act_2], axis=-1)
#
#         # to_torch
#         obs_1 = utils.to_torch(obs_1).to(self.device)
#         act_1 = utils.to_torch(act_1).to(self.device)
#         obs_2 = utils.to_torch(obs_2).to(self.device)
#         act_2 = utils.to_torch(act_2).to(self.device)
#
#         # get comparable labels
#         comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
#         comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)
#
#         # get step-wise rewards
#         r_hat1 = self.ensemble[member](obs_1, act_1)  # batch_size * traj_len
#         r_hat2 = self.ensemble[member](obs_2, act_2)
#
#         r_hat1_sum = r_hat1.sum(axis=1).unsqueeze(-1) # batch_size * 1
#         r_hat2_sum = r_hat2.sum(axis=1).unsqueeze(-1)
#
#         # should calculate before normalization for correct comparison
#         # r_hat = torch.cat([r_hat1, r_hat2], axis=2)
#         r_hat_sum = torch.cat([r_hat1_sum, r_hat2_sum], axis=1)  # batch_size * 2
#
#         # TODO how to normalize
#         # normalize to satisfy geometric mean strength = 1
#         # r_hat1 = (r_hat1 - r_hat.mean()) / (r_hat.std() + 1e-8)
#         # r_hat2 = (r_hat2 - r_hat.mean()) / (r_hat.std() + 1e-8)
#         r_hat1_sum = (r_hat1_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#         # r_hat2_sum = (r_hat2_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#
#
#         # calculate iterative objective
#         target_traj_reward = torch.zeros_like(r_hat1_sum)  # batch_size * 1
#         # target_step_reward = torch.zeros_like(r_hat1) # batch_size * traj_len * 1
#
#         # note that 'index' (i/j) denotes the index in the batch
#         # while 'start_index' denotes the starting timestep in the pref_data
#         unique_i_start_indices = np.unique(batch['start_indices'])
#         for start_index_i in unique_i_start_indices:
#
#             sum_1 = 0
#             sum_2 = 0
#
#             i = np.where(batch['start_indices'] == start_index_i)[0]
#
#             # avoid that the batch contains a trajectory many times (many identical i in the batch)
#             # they have the same start_indices but different batch index, calculate once is enough
#             if len(i) > 1:
#                 i = np.array([i[0]])
#
#             reward_i = r_hat1_sum[i].squeeze(-1)
#             exp_reward_i = torch.exp(reward_i)
#
#             start_index_j_list = []
#
#             # given an index i, only search j in the existing pair_labels (i,j,label) or (j,i,label)
#             for pair in win_lose_dict.keys():
#                 # order (i,j) '0'
#                 if pair[0] == start_index_i:
#                     start_index_j_list.append((pair[1], 0))
#                 # order (j,i) '1'
#                 elif pair[1] == start_index_i:
#                     start_index_j_list.append((pair[0], 1))
#
#             # automatically remove duplicates
#             unique_j_start_indices = set(start_index_j_list)
#
#             for (start_index_j, order) in unique_j_start_indices:
#                 obs_key = 'observations_2' if order == 0 else 'observations'
#                 act_key = 'actions_2' if order == 0 else 'actions'
#                 start_indices_key = 'start_indices_2' if order == 0 else 'start_indices'
#
#                 j = np.where(pref_dataset[start_indices_key] == start_index_j)[0]
#                 # avoid that the batch contains a trajectory many times (many identical j in the batch)
#                 # they have the same start_indices but different batch index, calculate once is enough
#                 if len(j) > 1:
#                     j = np.array([j[0]])
#
#                 obs_j = torch.from_numpy(pref_dataset[obs_key][j]).to(self.device).float()
#                 acs_j = torch.from_numpy(pref_dataset[act_key][j]).to(self.device).float()
#
#                 reward_j = self.ensemble[member](obs_j, acs_j).sum(axis=1).squeeze(-1)
#                 reward_j = (reward_j - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
#                 exp_reward_j = torch.exp(reward_j)
#
#                 w_ij, w_ji = get_winlose_times(win_lose_dict, start_index_i, start_index_j)
#
#                 sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
#                 sum_2 += w_ji / (exp_reward_i + exp_reward_j)
#             if self.risk_sensitive:
#                 if not self.beta_discrete:
#                     alpha, beta = self.beta_model.get_alpha_beta(torch.from_numpy(s_a_1[i]).float().to(self.device))
#                 else:
#                     alpha, beta = self.beta_model.get_rescaled_alpha_beta_discrete\
#                         (torch.from_numpy(s_a_1[i]).float().to(self.device), rescale=100)
#                 #print(f'alpha_{alpha.item()}_beta_{beta.item()}')
#                 target_traj_reward[i] = torch.log(((alpha/(exp_reward_i+1)) + sum_1) + 1e-8) - \
#                                         torch.log(((beta/(exp_reward_i+1)) + sum_2) + 1e-8)
#             else:
#                 target_traj_reward[i] = torch.log(((1/(exp_reward_i+1)) + sum_1) + 1e-8) - \
#                                         torch.log(((1/(exp_reward_i+1)) + sum_2) + 1e-8)
#
#         curr_loss = F.mse_loss(r_hat1_sum, target_traj_reward)
#         # print(curr_loss)
#
#
#         # compute acc
#         _, predicted = torch.max(r_hat_sum.data, 1)
#
#         if not len(comparable_indices):
#             correct = 0.7  # TODO, for exception
#         else:
#             correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
#                 comparable_indices)
#         return curr_loss, correct

class ItrTransformerRewardModel(MLPRewardModel):
    def __init__(self,
                 task, observation_dim, action_dim, structure_type="transformer",
                 ensemble_size=3, lr=0.0003, activation="tanh",
                 d_model=256, nhead=4, num_layers=1, max_seq_len=100,
                 logger=None, device="cuda", **kwargs):
        self.structure_type = structure_type
        self.d_model = d_model
        self.nhead = nhead
        self.num_layers = num_layers
        self.max_seq_len = max_seq_len
        self.reward_model_type = kwargs.get('reward_model_type', 'DistributionalTransformer')

        super().__init__(
            task, observation_dim, action_dim,
            ensemble_size, lr, activation, logger, device)

        self.risk_sensitive = kwargs.get('risk_sensitive', False)
        self.beta_model_path = kwargs.get('beta_model_path', None)
        self.beta_lr = kwargs.get('beta_lr', 3e-5)
        self.beta_n_epochs = kwargs.get('beta_n_epochs', 50)
        self.beta_model_type = kwargs.get('beta_model_type', 'Transformer')
        self.beta_coef = kwargs.get('beta_coeff', 0.01)
        self.beta_discrete = kwargs.get('beta_discrete', False)

    def construct_ensemble(self):
        for i in range(self.ensemble_size):
            if 'Distributional' in self.reward_model_type:
                transformer = utils.DistributionalPrefTransformer(
                    self.observation_dim, self.action_dim,
                    self.max_seq_len,
                    self.d_model, self.nhead, self.num_layers
                )
            else:
                transformer = utils.PrefTransformer1(
                    self.observation_dim, self.action_dim,
                    self.max_seq_len,
                    self.d_model, self.nhead, self.num_layers
                )
            self.ensemble.append(transformer.to(self.device))
            self.paramlst.extend(self.ensemble[-1].parameters())

        self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)

    def train(self, n_epochs, pref_dataset, data_size, batch_size):
        # preprocess pref_dataset to get win_lose_dict
        win_lose_dict = get_winlose_dict(pref_dataset)

        # train beta model
        if self.risk_sensitive:
            self.beta_model = utils.BetaNetwork(self.observation_dim, self.action_dim,
                                                self.beta_lr, self.device, self.beta_model_type, self.beta_coef)
            # continuous -- beta transformer
            if not self.beta_discrete:
                save_dir = Path(self.beta_model_path) / f'{self.task}' / f'{self.beta_model_type}_Coef{self.beta_coef}_lr{self.beta_lr}'
                model_dir = save_dir / f'itr_{self.beta_n_epochs}'
                model_dir.mkdir(parents=True, exist_ok=True)
                model_file = model_dir / 'beta_model.pth'
                if not model_file.is_file():
                    # train beta model
                    print('Start training Beta model.')
                    self.beta_model.fit_data(num_epochs=self.beta_n_epochs, data_size=data_size,
                                             batch_size=batch_size, dataset=pref_dataset, save_dir=save_dir)
                    print('Finish training Beta model.')
                    print(f'Save Beta model in {save_dir}.')
                    print('PLease rerun the code to load the trained Beta model.')
                    exit()
                    # self.beta_model.save_model(model_file)
                else:
                    # load beta model
                    print(f'Load Beta model from {model_file}.')
                    self.beta_model.load_model(model_file)
            # discrete -- beta dict
            else:
                save_dir = Path(self.beta_model_path) / f'{self.task}' / 'discrete'
                save_dir.mkdir(parents=True, exist_ok=True)
                model_file = save_dir / 'discrete_beta_dict.pkl'
                if not model_file.is_file():
                    print('Start learning Beta dict.')
                    self.beta_model.fit_data_discrete(dataset=pref_dataset)
                    print('Finish learning Beta dict.')
                    print(f'Save Beta model in {save_dir}.')
                    self.beta_model.save_model(model_file)
                    with open(model_file, 'wb') as pickle_file:
                        pickle.dump(self.beta_model.traj_alpha_beta_dict, pickle_file)
                    print('PLease rerun the code to load the trained Beta model.')
                    exit()
                else:
                    # load beta model
                    print(f'Load Beta model from {model_file}.')
                    with open(model_file, 'rb') as pickle_file:
                        beta_dict = pickle.load(pickle_file)
                    self.beta_model.traj_alpha_beta_dict = beta_dict

        # train reward model
        print('Start training reward model.')
        interval = int(data_size / batch_size) + 1

        for epoch in range(1, n_epochs + 1):
            ensemble_losses = [[] for _ in range(self.ensemble_size)]
            ensemble_acc = [[] for _ in range(self.ensemble_size)]

            batch_shuffled_idx = []
            for _ in range(self.ensemble_size):
                batch_shuffled_idx.append(np.random.permutation(pref_dataset["observations"].shape[0]))

            self.target_traj_reward_dict = {}

            for i in tqdm(range(interval)):
                self.opt.zero_grad()
                total_loss = 0
                start_pt = i * batch_size
                end_pt = min((i + 1) * batch_size, pref_dataset["observations"].shape[0])
                for member in range(self.ensemble_size):
                    # get batch
                    batch = index_batch(pref_dataset, batch_shuffled_idx[member][start_pt:end_pt])
                    # compute loss
                    win_lose_dict = get_winlose_dict(batch)
                    curr_loss, correct = self._train(win_lose_dict, batch, member)
                    total_loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    ensemble_acc[member].append(correct)
                total_loss.backward()
                self.opt.step()

            train_metrics = {"epoch": epoch,
                             "avg_loss": np.mean(ensemble_losses),
                             "avg_acc": np.mean(ensemble_acc)}
            for i in range(self.ensemble_size):
                train_metrics.update({f"ensemble_{i}_loss": np.mean(ensemble_losses[i])})
                train_metrics.update({f"ensemble_{i}_acc": np.mean(ensemble_acc[i])})
            self.logger.log(train_metrics)

            # early stop
            if np.mean(ensemble_acc) > 0.99 and "antmaze" not in self.task:
                break

            # save per 10 epoch
            if (epoch-1) % 10 == 0:
                dir_path = self.logger._model_dir / f'epoch_{epoch-1}'
                try:
                    os.makedirs(dir_path)
                except OSError:
                    pass
                self.save_model(dir_path / f"reward_model.pt")

        print('Finish training reward model.')

    def _train(self,win_lose_dict, batch, member):
        # TODO
        # maybe need to handle with tie in the future (now I just skip these pairs)

        # get batch
        obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
        act_1 = batch['actions']  # batch_size * traj_len * action_dim
        obs_2 = batch['observations_2']
        act_2 = batch['actions_2']
        labels = batch['labels']  # batch_size * 2 (one-hot, for equal label)
        s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
        s_a_2 = np.concatenate([obs_2, act_2], axis=-1)

        # to_torch
        obs_1 = utils.to_torch(obs_1).to(self.device)
        act_1 = utils.to_torch(act_1).to(self.device)
        obs_2 = utils.to_torch(obs_2).to(self.device)
        act_2 = utils.to_torch(act_2).to(self.device)

        # get comparable labels
        comparable_indices = np.where((labels != [0.5, 0.5]).any(axis=1))[0]
        comparable_labels = torch.from_numpy(np.argmax(labels, axis=1)).to(self.device)

        # get step-wise rewards
        r_hat1 = self.ensemble[member](obs_1, act_1)  # batch_size * traj_len
        r_hat2 = self.ensemble[member](obs_2, act_2)

        r_hat1_sum = r_hat1.sum(axis=1).unsqueeze(-1) # batch_size * 1
        r_hat2_sum = r_hat2.sum(axis=1).unsqueeze(-1)

        # should calculate before normalization for correct comparison
        # r_hat = torch.cat([r_hat1, r_hat2], axis=2)
        r_hat_sum = torch.cat([r_hat1_sum, r_hat2_sum], axis=1)  # batch_size * 2

        # TODO how to normalize do not normalize, just force mean
        # normalize to satisfy geometric mean strength = 1
        # r_hat1 = (r_hat1 - r_hat.mean()) / (r_hat.std() + 1e-8)
        # r_hat2 = (r_hat2 - r_hat.mean()) / (r_hat.std() + 1e-8)
        r_hat1_sum = (r_hat1_sum - r_hat_sum.mean())
        r_hat2_sum = (r_hat2_sum - r_hat_sum.mean())
        # r_hat1_sum = (r_hat1_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)
        # r_hat2_sum = (r_hat2_sum - r_hat_sum.mean()) / (r_hat_sum.std() + 1e-8)


        # calculate iterative objective
        target_traj_reward = torch.zeros_like(r_hat1_sum)  # batch_size * 1
        # target_step_reward = torch.zeros_like(r_hat1) # batch_size * traj_len * 1

        # note that 'index' (i/j) denotes the index in the batch
        # while 'start_index' denotes the starting timestep in the pref_data
        unique_i_start_indices = np.unique(batch['start_indices'])
        for start_index_i in unique_i_start_indices:

            sum_1 = 0
            sum_2 = 0

            i = np.where(batch['start_indices'] == start_index_i)[0]

            # avoid that the batch contains a trajectory many times (many identical i in the batch)
            # they have the same start_indices but different batch index, calculate once is enough
            if len(i) > 1:
                i = np.array([i[0]])

            reward_i = r_hat1_sum[i].squeeze(-1)
            exp_reward_i = torch.exp(reward_i)

            start_index_j_list = []

            # given an index i, only search j in the existing pair_labels (i,j,label) or (j,i,label)
            for pair in win_lose_dict.keys():
                # order (i,j) '0'
                if pair[0] == start_index_i:
                    start_index_j_list.append((pair[1], 0))
                # order (j,i) '1'
                elif pair[1] == start_index_i:
                    start_index_j_list.append((pair[0], 1))

            # automatically remove duplicates
            unique_j_start_indices = set(start_index_j_list)

            for (start_index_j, order) in unique_j_start_indices:
                start_indices_key = 'start_indices_2' if order == 0 else 'start_indices'

                j = np.where(batch[start_indices_key] == start_index_j)[0]
                # avoid that the batch contains a trajectory many times (many identical j in the batch)
                # they have the same start_indices but different batch index, calculate once is enough
                if len(j) > 1:
                    j = np.array([j[0]])

                reward_j = r_hat2_sum[j] if order == 0 else r_hat1_sum[j]
                exp_reward_j = torch.exp(reward_j)

                w_ij, w_ji = get_winlose_times(win_lose_dict, start_index_i, start_index_j)

                sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
                sum_2 += w_ji / (exp_reward_i + exp_reward_j)

            if self.risk_sensitive:
                if not self.beta_discrete:
                    alpha, beta = self.beta_model.get_alpha_beta(torch.from_numpy(s_a_1[i]).float().to(self.device),rescale=10)
                else:
                    alpha, beta = self.beta_model.get_rescaled_alpha_beta_discrete\
                        (torch.from_numpy(s_a_1[i]).float().to(self.device), rescale=100)
                #print(f'alpha_{alpha.item()}_beta_{beta.item()}')
                # TODO, modify when k!=1
                with torch.no_grad():
                    for k in range(1):
                        reward_i = torch.log(torch.clamp(((alpha/(exp_reward_i+1)) + sum_1), min=1e-8)) - \
                                            torch.log(torch.clamp(((beta/(exp_reward_i+1)) + sum_2), min=1e-8))
                        # reward_i = torch.log(torch.clamp((sum_1), min=1e-8)) - \
                        #                     torch.log(torch.clamp((sum_2), min=1e-8))
                        exp_reward_i_next = torch.exp(reward_i)
                        delta = abs(exp_reward_i_next - exp_reward_i)
                        exp_reward_i = exp_reward_i_next
                        if delta < 1e-3:
                            break
                    # target_traj_reward[i] = exp_reward_i
                    self.target_traj_reward_dict = self.update_target_trajectory_dict\
                        (self.target_traj_reward_dict, start_index_i, reward_i)
                    reward_list = self.target_traj_reward_dict.get(start_index_i, [])
                    target_traj_reward[i] = random.choice(reward_list)
            else:
                with torch.no_grad():
                    for k in range(1):
                        sum_1 += w_ij * exp_reward_j / (exp_reward_i + exp_reward_j)
                        sum_2 += w_ji / (exp_reward_i + exp_reward_j)
                        reward_i = torch.log(torch.clamp(((1/(exp_reward_i+1)) + sum_1), min=1e-8)) - \
                                            torch.log(torch.clamp(((1/(exp_reward_i+1)) + sum_2), min=1e-8))
                        # reward_i = torch.log(torch.clamp((sum_1), min=1e-8)) - \
                        #                     torch.log(torch.clamp((sum_2), min=1e-8))
                        exp_reward_i_next = torch.exp(reward_i)
                        delta = abs(exp_reward_i_next - exp_reward_i)
                        exp_reward_i = exp_reward_i_next
                        if delta < 1e-3:
                            break
                    # target_traj_reward[i] = exp_reward_i
                    self.target_traj_reward_dict = self.update_target_trajectory_dict\
                        (self.target_traj_reward_dict, start_index_i, reward_i)
                    reward_list = self.target_traj_reward_dict.get(start_index_i, [])
                    target_traj_reward[i] = random.choice(reward_list)

        curr_loss = F.mse_loss(r_hat1_sum, target_traj_reward)
        # print(curr_loss)


        # compute acc
        _, predicted = torch.max(r_hat_sum.data, 1)

        if not len(comparable_indices):
            correct = 0.7  # TODO, for exception
        else:
            correct = (predicted[comparable_indices] == comparable_labels[comparable_indices]).sum().item() / len(
                comparable_indices)
        return curr_loss, correct

    def update_target_trajectory_dict(self, dict, trajectory, target_reward):
        if trajectory in dict:
            dict[trajectory].append(target_reward)
        else:
            dict[trajectory] = [target_reward]
        return dict