import collections
import numpy as np
import gym
from tqdm import trange
import torch
import torch.nn as nn
import math
import d4rl
from tqdm import tqdm
import torch.nn.functional as F


Batch = collections.namedtuple(
    'Batch',
    ['observations', 'actions', 'rewards', 'masks', 'next_observations'])


def gen_net(in_size=1, out_size=1, H=128, n_layers=3, activation='tanh'):
    net = []
    for i in range(n_layers):
        net.append(nn.Linear(in_size, H))
        net.append(nn.LeakyReLU())
        in_size = H
    net.append(nn.Linear(in_size, out_size))
    if activation == 'tanh':
        net.append(nn.Tanh())
    elif activation == 'sig':
        net.append(nn.Sigmoid())
    else:
        pass

    return net

def get_discrete_traj(trajectory, dim=2):
    discrete_traj = np.round(trajectory[:, :, :dim].astype(float)).astype(int)

    unique_states_per_trajectory = []
    for trajectory in discrete_traj:
        # Find unique states in the current trajectory
        unique_states = np.unique(trajectory, axis=0)
        # Convert the trajectory to a hashable type
        unique_states_per_trajectory.append(tuple(map(tuple, unique_states)))
    return unique_states_per_trajectory

def init_trajectory_dict(discrete_trajectory):

    trajectory_dict = {}
    for trajectory in discrete_trajectory:
        if trajectory in trajectory_dict:
            continue
        else:
            trajectory_dict[trajectory] = np.array([1,1])
    return trajectory_dict

def get_trajectory_dict_from_pair(traj_alpha_beta_dict, discrete_obs_1, discrete_obs_2, single_labels):
    for i in range(len(single_labels)):
        if single_labels[i] == 1:
            traj_alpha_beta_dict[discrete_obs_1[i]][0] += 1
            traj_alpha_beta_dict[discrete_obs_2[i]][1] += 1
        elif single_labels[i] == -1:
            traj_alpha_beta_dict[discrete_obs_1[i]][1] += 1
            traj_alpha_beta_dict[discrete_obs_2[i]][0] += 1
        # TODO how to treat not compaied ones?
        elif single_labels[i] == 0:
            traj_alpha_beta_dict[discrete_obs_1[i]][0] += 1
            traj_alpha_beta_dict[discrete_obs_1[i]][1] += 1
            traj_alpha_beta_dict[discrete_obs_2[i]][0] += 1
            traj_alpha_beta_dict[discrete_obs_2[i]][1] += 1
    return traj_alpha_beta_dict

def to_torch(x, dtype=torch.float32):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).to(dtype)

def index_batch(batch, indices):
    indexed = {}
    for key in batch.keys():
        indexed[key] = batch[key][indices, ...]
    return indexed

class Dataset(object):
    def __init__(self, observations: np.ndarray, actions: np.ndarray,
                 rewards: np.ndarray, masks: np.ndarray,
                 dones_float: np.ndarray, next_observations: np.ndarray,
                 size: int):
        self.observations = observations
        self.actions = actions
        self.rewards = rewards
        self.masks = masks
        self.dones_float = dones_float
        self.next_observations = next_observations
        self.size = size

    def sample(self, batch_size: int) -> Batch:
        indx = np.random.randint(self.size, size=batch_size)
        return Batch(observations=self.observations[indx],
                     actions=self.actions[indx],
                     rewards=self.rewards[indx],
                     masks=self.masks[indx],
                     next_observations=self.next_observations[indx])


class D4RLDataset(Dataset):
    def __init__(self,
                 env: gym.Env,
                 clip_to_eps: bool = True,
                 eps: float = 1e-5):
        dataset = d4rl.qlearning_dataset(env)

        if clip_to_eps:
            lim = 1 - eps
            dataset['actions'] = np.clip(dataset['actions'], -lim, lim)

        dones_float = np.zeros_like(dataset['rewards'])

        for i in range(len(dones_float) - 1):
            if np.linalg.norm(dataset['observations'][i + 1] -
                              dataset['next_observations'][i]
                              ) > 1e-5 or dataset['terminals'][i] == 1.0:
                dones_float[i] = 1
            else:
                dones_float[i] = 0

        dones_float[-1] = 1

        super().__init__(dataset['observations'].astype(np.float32),
                         actions=dataset['actions'].astype(np.float32),
                         rewards=dataset['rewards'].astype(np.float32),
                         masks=1.0 - dataset['terminals'].astype(np.float32),
                         dones_float=dones_float.astype(np.float32),
                         next_observations=dataset['next_observations'].astype(
                             np.float32),
                         size=len(dataset['observations']))


@torch.no_grad()
def reward_from_preference(
    dataset: D4RLDataset,
    reward_model,
    batch_size: int = 256,
    reward_model_type: str = "transformer",
    device="cuda"
):
    data_size = dataset["rewards"].shape[0]
    interval = int(data_size / batch_size) + 1
    new_r = np.zeros_like(dataset["rewards"])
    
    if reward_model_type == "transformer":
        max_seq_len = reward_model.max_seq_len
        for each in reward_model.ensemble:
            each.eval()
 
        obs, act = [], []
        ptr = 0
        for i in trange(data_size):
            
            if len(obs) < max_seq_len:
                obs.append(dataset["observations"][i])
                act.append(dataset["actions"][i])
            
            if dataset["terminals"][i] > 0 or i == data_size - 1 or len(obs) == max_seq_len:
                tensor_obs = to_torch(np.array(obs)[None,], dtype=torch.float32).to(device)
                tensor_act = to_torch(np.array(act)[None,], dtype=torch.float32).to(device)
                
                new_reward = 0
                for each in reward_model.ensemble:
                    new_reward += each(tensor_obs, tensor_act).detach().cpu().numpy()
                new_reward /= len(reward_model.ensemble)
                if tensor_obs.shape[1] <= -1:
                    new_r[ptr:ptr+tensor_obs.shape[1]] = dataset["rewards"][ptr:ptr+tensor_obs.shape[1]]
                else:
                    new_r[ptr:ptr+tensor_obs.shape[1]] = new_reward
                ptr += tensor_obs.shape[1]
                obs, act = [], []
    else:
        for i in trange(interval):
            start_pt = i * batch_size
            end_pt = (i + 1) * batch_size

            observations = dataset["observations"][start_pt:end_pt]
            actions = dataset["actions"][start_pt:end_pt]
            obs_act = np.concatenate([observations, actions], axis=-1)

            new_reward = reward_model.get_reward_batch(obs_act).reshape(-1)
            new_r[start_pt:end_pt] = new_reward
    
    dataset["rewards"] = new_r.copy()
    
    # rr = dataset["rewards"].copy()
    # fr = new_r.copy()
    
    # rr = (rr-rr.min())/(rr.max()-rr.min())
    # fr = (fr-fr.min())/(fr.max()-fr.min())
    
    # rr_n_bins, _ = np.histogram(rr, 10, (0, 1))
    # fr_n_bins, _ = np.histogram(fr, 10, (0, 1))
    
    # print(rr_n_bins)
    # print(fr_n_bins)
    
    return dataset

BETA = 1e-8
def reward_from_preference_ensemble(
        dataset: D4RLDataset,
        reward_model,
        batch_size: int = 256,
        reward_model_type: str = "transformer",
        device="cuda"
):
    data_size = dataset["rewards"].shape[0]
    interval = int(data_size / batch_size) + 1
    new_r = np.zeros_like(dataset["rewards"])

    if reward_model_type == "transformer":
        max_seq_len = reward_model.max_seq_len
        for each in reward_model.ensemble:
            each.eval()

        obs, act = [], []
        ptr = 0
        for i in trange(data_size):

            if len(obs) < max_seq_len:
                obs.append(dataset["observations"][i])
                act.append(dataset["actions"][i])

            if dataset["terminals"][i] > 0 or i == data_size - 1 or len(obs) == max_seq_len:
                tensor_obs = to_torch(np.array(obs)[None,], dtype=torch.float32).to(device)
                tensor_act = to_torch(np.array(act)[None,], dtype=torch.float32).to(device)

                rewards_list = []
                for each in reward_model.ensemble:
                    rewards_list.append(each(tensor_obs, tensor_act).detach().cpu().numpy())
                rewards_array = np.array(rewards_list)
                mean_reward = np.mean(rewards_array, axis=0)
                std_reward = np.std(rewards_array, axis=0)
                new_reward = mean_reward - BETA * std_reward
                if tensor_obs.shape[1] <= -1:
                    new_r[ptr:ptr + tensor_obs.shape[1]] = dataset["rewards"][ptr:ptr + tensor_obs.shape[1]]
                else:
                    new_r[ptr:ptr + tensor_obs.shape[1]] = new_reward
                ptr += tensor_obs.shape[1]
                obs, act = [], []
    else:
        for i in trange(interval):
            start_pt = i * batch_size
            end_pt = (i + 1) * batch_size

            observations = dataset["observations"][start_pt:end_pt]
            actions = dataset["actions"][start_pt:end_pt]
            obs_act = np.concatenate([observations, actions], axis=-1)

            new_reward = reward_model.get_reward_batch(obs_act).reshape(-1)
            new_r[start_pt:end_pt] = new_reward

    dataset["rewards"] = new_r.copy()

    # rr = dataset["rewards"].copy()
    # fr = new_r.copy()

    # rr = (rr-rr.min())/(rr.max()-rr.min())
    # fr = (fr-fr.min())/(fr.max()-fr.min())

    # rr_n_bins, _ = np.histogram(rr, 10, (0, 1))
    # fr_n_bins, _ = np.histogram(fr, 10, (0, 1))

    # print(rr_n_bins)
    # print(fr_n_bins)

    return dataset

def dirichlet_kl_divergence_loss(alpha, prior):
    """
    KL divergence between two dirichlet distribution
    The mean is alpha/(alpha+beta) and variance is alpha*beta/(alpha+beta)^2*(alpha+beta+1)
    There are multiple ways of modelling a dirichlet:
    1) by Laplace approximation with logistic normal: https://arxiv.org/pdf/1703.01488.pdf
    2) by directly modelling dirichlet parameters: https://arxiv.org/pdf/1901.02739.pdf
    code reference：
    1） https://github.com/sophieburkhardt/dirichlet-vae-topic-models
    2） https://github.com/is0383kk/Dirichlet-VAE
    """
    analytical_kld = torch.lgamma(torch.sum(alpha, dim=1)) - torch.lgamma(torch.sum(prior, dim=1))
    analytical_kld += torch.sum(torch.lgamma(prior), dim=1)
    analytical_kld -= torch.sum(torch.lgamma(alpha), dim=1)
    minus_term = alpha - prior
    # tmp = torch.reshape(torch.digamma(torch.sum(alpha, dim=1)), shape=[alpha.shape[0], 1])
    digamma_term = torch.digamma(alpha) - \
                   torch.reshape(torch.digamma(torch.sum(alpha, dim=1)), shape=[alpha.shape[0], 1])
    test = torch.sum(torch.mul(minus_term, digamma_term), dim=1)
    analytical_kld += test
    # self.analytical_kld = self.mask * self.analytical_kld  # mask paddings
    return analytical_kld


class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        device = x.device
        half_dim = self.dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=device) * -emb)
        emb = x[:, None] * emb[None, :]
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb

class TransformerBetaModel(nn.Module):
    def __init__(self, data_dim, embedding_dim, nhead, num_encoder_layers, output_dim, device):
        super(TransformerBetaModel, self).__init__()
        self.device = device
        self.embedding = nn.Linear(data_dim, embedding_dim)
        # self.pos_encoder = self.create_positional_encoding(seq_length, embedding_dim)
        self.pos_emb = SinusoidalPosEmb(embedding_dim)
        encoder_layers = nn.TransformerEncoderLayer(d_model=embedding_dim, nhead=nhead, batch_first=True)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, num_layers=num_encoder_layers)
        self.output_layer = nn.Linear(embedding_dim, output_dim)
        self.softplus = nn.Softplus()

    def forward(self, x):
        traj_len = x.shape[1]
        x = self.embedding(x)  # Map input to embedding dimension
        pos = self.pos_emb(
            torch.arange(traj_len, device=self.device))[None,]
        x += pos  # Add positional encoding
        output = self.transformer_encoder(x)
        output = self.output_layer(output)
        # TODO mean or last one
        output = output.mean(dim=1)  # Aggregate across the sequence dimension
        # output = output[:, -1, :]  # Take the output from the last timestep
        output = self.softplus(output)
        return output

class CausalTransformerBetaModel(nn.Module):
    def __init__(self, data_dim, embedding_dim, nhead, num_encoder_layers, output_dim, device):
        super(CausalTransformerBetaModel, self).__init__()
        self.device = device
        self.embedding = nn.Linear(data_dim, embedding_dim)
        # self.pos_encoder = self.create_positional_encoding(seq_length, embedding_dim)
        self.pos_emb = SinusoidalPosEmb(embedding_dim)
        encoder_layers = nn.TransformerEncoderLayer(d_model=embedding_dim, nhead=nhead, batch_first=True)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, num_layers=num_encoder_layers)
        self.output_layer = nn.Linear(embedding_dim, output_dim)
        self.softplus = nn.Softplus()

    def generate_square_subsequent_mask(self, sz):
        mask = torch.triu(torch.ones(sz, sz), diagonal=1)
        return mask.masked_fill(mask==1, float('-inf'))

    def forward(self, x):
        traj_len = x.shape[1]
        x = self.embedding(x)  # Map input to embedding dimension
        pos = self.pos_emb(
            torch.arange(traj_len, device=self.device))[None,]
        x += pos  # Add positional encoding
        mask = self.generate_square_subsequent_mask(traj_len).to(x.device)
        output = self.transformer_encoder(x, mask=mask)
        output = self.output_layer(output)
        # TODO mean or last one
        # output = output.mean(dim=1)  # Aggregate across the sequence dimension
        output = output[:, -1, :]  # Take the output from the last timestep
        output = self.softplus(output)
        return output

class MLPBetaModel(nn.Module):
    def __init__(self, data_dim, device):
        super(MLPBetaModel, self).__init__()
        self.device = device
        self.network = nn.Sequential(*gen_net(in_size=data_dim,
                                            out_size=2, H=64, n_layers=3,
                                            activation=None)).float().to(self.device)
        # self.backbone_net = nn.Sequential(
        #     nn.Linear(data_dim, 64),
        #     nn.LeakyReLU(),
        #     nn.Linear(64, 256),
        #     nn.LeakyReLU(),
        #     nn.Linear(256, 256),
        #     nn.LeakyReLU(),
        # ).to(self.device)
        #
        # self.alpha_net = nn.Sequential(
        #     nn.Linear(256, 64),
        #     nn.LeakyReLU(),
        #     nn.Linear(64, 64),
        #     nn.LeakyReLU(),
        #     nn.Linear(64, 1),  # Assuming alpha is a single value
        # ).to(self.device)
        #
        # self.beta_net = nn.Sequential(
        #     nn.Linear(256, 64),
        #     nn.LeakyReLU(),
        #     nn.Linear(64, 64),
        #     nn.LeakyReLU(),
        #     nn.Linear(64, 1),  # Assuming alpha is a single value
        # ).to(self.device)
        self.activation = nn.LeakyReLU()
        self.softplus = nn.Softplus()

    def forward(self, x):
        output = self.network(x)
        output = output.mean(dim=1)  # Aggregate across the sequence dimension
        output = self.softplus(output)

        # output = self.backbone_net(x)
        # alpha = self.softplus(self.alpha_net(output).mean(dim=1)).squeeze(-1)
        # beta = self.softplus(self.beta_net(output).mean(dim=1)).squeeze(-1)
        # return alpha, beta
        return output

class BetaNetwork(nn.Module):
    def __init__(self,observation_dim, action_dim, lr=3e-4, device=torch.device('cuda'),
                 model_type='Transformer', beta_coef=0.1):
        super(BetaNetwork, self).__init__()
        self.observation_dim = observation_dim
        self.action_dim = action_dim
        self.lr = lr
        self.beta_coef = beta_coef
        self.device = device

        if 'Causal' in model_type:
            self.model = CausalTransformerBetaModel(
                data_dim=self.observation_dim + self.action_dim,
                embedding_dim=256,
                nhead=4,
                num_encoder_layers=1,
                output_dim=2,
                device=self.device,
            ).to(self.device)
        elif 'Transformer' in model_type:
            self.model = TransformerBetaModel(
                data_dim = self.observation_dim + self.action_dim,
                embedding_dim = 256,
                nhead = 4,
                num_encoder_layers = 1,
                output_dim = 2,
                device = self.device,
            ).to(self.device)
        else:
            self.model = MLPBetaModel(
                data_dim = self.observation_dim + self.action_dim,
                device = self.device,
            )

        self.opt = torch.optim.Adam(self.model.parameters(), lr=self.lr)

    def get_alpha_beta(self, x, rescale=1):
        alpha_beta = self.model(x).detach()
        alpha = alpha_beta[:, 0].detach()
        beta = alpha_beta[:, 1].detach()
        # alpha, beta = self.model(x)
        alpha_normalized =torch.max(alpha / rescale, torch.tensor(1.0))
        beta_normalized = torch.max(beta / rescale, torch.tensor(1.0))
        return alpha_normalized, beta_normalized

    def kl_regularizer_loss(self, batch_size, alpha, beta):
        prior = torch.tensor(np.asarray(batch_size * [[1, 1]]), dtype=torch.float32).to(self.device)
        analytical_kld_loss = dirichlet_kl_divergence_loss(
            alpha=torch.stack([alpha, beta], dim=1),
            prior=prior).mean()
        return analytical_kld_loss

    # def fit_data(self, num_epochs, data_size, batch_size, dataset, save_dir):
    #     interval = int(data_size / batch_size) + 1
    #     for epoch in range(num_epochs):
    #
    #         beta_loss_1 = []
    #         beta_loss_2 = []
    #         beta_loss_all = []
    #
    #         batch_shuffled_idx = np.random.permutation(dataset["observations"].shape[0])
    #         for i in tqdm(range(interval)):
    #
    #             start_pt = i * batch_size
    #             end_pt = min((i + 1) * batch_size, dataset["observations"].shape[0])
    #             batch = index_batch(dataset, batch_shuffled_idx[start_pt:end_pt])
    #
    #             # get batch
    #             obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
    #             act_1 = batch['actions']  # batch_size * traj_len * action_dim
    #             obs_2 = batch['observations_2']
    #             act_2 = batch['actions_2']
    #             s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
    #             s_a_2 = np.concatenate([obs_2, act_2], axis=-1)
    #
    #             conditions = [np.all(batch['labels'] == [1, 0], axis=1),
    #                           np.all(batch['labels'] == [0, 1], axis=1),
    #                           np.all(batch['labels'] == [0.5, 0.5], axis=1)]
    #             values = [1, -1, 0]
    #             single_labels = torch.from_numpy(np.select(conditions, values)).float().to(self.device)
    #
    #
    #             # pred_1_alpha_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device)) # batch_size * 2
    #             # pred_1_alpha = pred_1_alpha_beta[:, 0]
    #             # pred_1_beta = pred_1_alpha_beta[:, 1]
    #             # pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #             #
    #             #
    #             # pred_2_alpha_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #             # pred_2_alpha = pred_2_alpha_beta[:, 0]
    #             # pred_2_beta = pred_2_alpha_beta[:, 1]
    #             # pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #
    #             pred_1_alpha, pred_1_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device))
    #             pred_2_alpha, pred_2_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #
    #             pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #             pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #
    #             # if equal, then discard
    #             # TODO maybe if equal, then both towards 1 (both preferred)
    #             loss_1 = torch.mean(torch.log(pred_1_preds) * single_labels) - torch.mean(torch.log(pred_2_preds) * single_labels)
    #             loss_2 = self.kl_regularizer_loss(pred_1_preds.shape[0], alpha=pred_1_alpha, beta=pred_1_beta) + \
    #                      self.kl_regularizer_loss(pred_2_preds.shape[0], alpha=pred_2_alpha, beta=pred_2_beta)
    #
    #             beta_loss = -loss_1 + self.beta_coef * loss_2
    #
    #             beta_loss_1.append(-loss_1)
    #             beta_loss_2.append(loss_2)
    #             beta_loss_all.append(beta_loss)
    #
    #             self.opt.zero_grad()
    #             beta_loss.backward()
    #             self.opt.step()
    #
    #         beta_loss_1 = torch.stack(beta_loss_1, dim=0)
    #         beta_loss_2 = torch.stack(beta_loss_2, dim=0)
    #         beta_loss_all = torch.stack(beta_loss_all, dim=0)
    #         print("iteration:", epoch + 1)
    #         print("mean_beta_loss_data:", torch.mean(beta_loss_1).item())
    #         print("mean_beta_loss_kl:", torch.mean(beta_loss_2).item())
    #         print("mean_beta_loss_all:", torch.mean(beta_loss_all).item())
    #
    #         if (epoch+1) % 10 == 0:
    #             tmp_save_dir= save_dir / f'itr_{epoch+1}'
    #             tmp_save_dir.mkdir(parents=True, exist_ok=True)
    #             model_file = tmp_save_dir / 'beta_model.pth'
    #             self.save_model(model_file)
    #
    # def fit_data(self, num_epochs, data_size, batch_size, dataset, save_dir):
    #     interval = int(data_size / batch_size) + 1
    #     for epoch in range(num_epochs):
    #
    #         beta_loss_1 = []
    #         beta_loss_2 = []
    #         beta_loss_all = []
    #
    #         batch_shuffled_idx = np.random.permutation(dataset["observations"].shape[0])
    #         for i in tqdm(range(interval)):
    #
    #             start_pt = i * batch_size
    #             end_pt = min((i + 1) * batch_size, dataset["observations"].shape[0])
    #             batch = index_batch(dataset, batch_shuffled_idx[start_pt:end_pt])
    #
    #             # get batch
    #             obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
    #             act_1 = batch['actions']  # batch_size * traj_len * action_dim
    #             obs_2 = batch['observations_2']
    #             act_2 = batch['actions_2']
    #             s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
    #             s_a_2 = np.concatenate([obs_2, act_2], axis=-1)
    #
    #             conditions = [np.all(batch['labels'] == [1, 0], axis=1),
    #                           np.all(batch['labels'] == [0, 1], axis=1),
    #                           np.all(batch['labels'] == [0.5, 0.5], axis=1)]
    #             values = [1, -1, 0]
    #             single_labels = torch.from_numpy(np.select(conditions, values)).float().to(self.device)
    #
    #
    #             pred_1_alpha_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device)) # batch_size * 2
    #             pred_1_alpha = pred_1_alpha_beta[:, 0]
    #             pred_1_beta = pred_1_alpha_beta[:, 1]
    #             pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #
    #
    #             pred_2_alpha_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #             pred_2_alpha = pred_2_alpha_beta[:, 0]
    #             pred_2_beta = pred_2_alpha_beta[:, 1]
    #             pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #
    #             # pred_1_alpha, pred_1_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device))
    #             # pred_2_alpha, pred_2_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #
    #             # if equal, then discard
    #             # TODO maybe if equal, then both towards 1 (both preferred)
    #             # loss_1 = torch.mean(pred_1_alpha * single_labels) + torch.mean(pred_2_alpha * single_labels)
    #             # for i in range(len(single_labels)):
    #             #     if single_labels[i] == 1:
    #             #         pred_1_beta[i].detach()
    #             #         pred_2_alpha[i].detach()
    #             #     elif single_labels[i] == -1:
    #             #         pred_1_alpha[i].detach()
    #             #         pred_2_beta[i].detach()
    #
    #             # Apply similar logic based on your conditions for when to update beta or alpha
    #             # Detach and calculate as needed
    #
    #             # pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #             # pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #
    #             # # loss_1 = torch.mean(torch.log(pred_1_preds) * single_labels) - torch.mean(torch.log(pred_2_preds) * single_labels)
    #             # loss_1 = torch.mean((pred_1_preds) * single_labels) - torch.mean((pred_2_preds) * single_labels)
    #             # loss_2 = self.kl_regularizer_loss(pred_1_alpha, alpha=pred_1_alpha, beta=pred_1_beta) + \
    #             #          self.kl_regularizer_loss(pred_2_alpha, alpha=pred_2_alpha, beta=pred_2_beta)
    #             #
    #             # beta_loss = -loss_1 + self.beta_coef * loss_2
    #             #
    #             # beta_loss_1.append(-loss_1)
    #             # beta_loss_2.append(loss_2)
    #             # beta_loss_all.append(beta_loss)
    #             #
    #             # self.opt.zero_grad()
    #             # beta_loss.backward()
    #             # self.opt.step()
    #             target_pred_1_alpha = pred_1_alpha.detach() + (single_labels == 1).float()
    #             target_pred_1_beta = pred_1_beta.detach() + (single_labels == -1).float()
    #
    #             target_pred_2_alpha = pred_2_alpha.detach() + (single_labels == -1).float()
    #             target_pred_2_beta = pred_2_beta.detach() + (single_labels == 1).float()
    #
    #
    #             loss_alpha_1 = F.mse_loss(pred_1_alpha, target_pred_1_alpha)
    #             loss_beta_1 = F.mse_loss(pred_1_beta, target_pred_1_beta)
    #
    #             loss_alpha_2 = F.mse_loss(pred_2_alpha, target_pred_2_alpha)
    #             loss_beta_2 = F.mse_loss(pred_2_beta, target_pred_2_beta)
    #             loss_1 = 0.5*loss_alpha_1 + 0.5*loss_beta_1
    #             # loss_2 = self.kl_regularizer_loss(pred_1_alpha, alpha=pred_1_alpha, beta=pred_1_beta) + \
    #             #          self.kl_regularizer_loss(pred_2_alpha, alpha=pred_2_alpha, beta=pred_2_beta)
    #
    #             beta_loss = loss_1
    #
    #             beta_loss_1.append(loss_1)
    #             # beta_loss_2.append(loss_2)
    #             beta_loss_all.append(beta_loss)
    #
    #             self.opt.zero_grad()
    #             beta_loss.backward()
    #             self.opt.step()
    #
    #         beta_loss_1 = torch.stack(beta_loss_1, dim=0)
    #         # beta_loss_2 = torch.stack(beta_loss_2, dim=0)
    #         beta_loss_all = torch.stack(beta_loss_all, dim=0)
    #         print("iteration:", epoch + 1)
    #         print("mean_beta_loss_data:", torch.mean(beta_loss_1).item())
    #         # print("mean_beta_loss_kl:", torch.mean(beta_loss_2).item())
    #         print("mean_beta_loss_all:", torch.mean(beta_loss_all).item())
    #
    #         if (epoch+1) % 10 == 0:
    #             tmp_save_dir= save_dir / f'itr_{epoch+1}'
    #             tmp_save_dir.mkdir(parents=True, exist_ok=True)
    #             model_file = tmp_save_dir / 'beta_model.pth'
    #             self.save_model(model_file)

    # def fit_data(self, num_epochs, data_size, batch_size, dataset, save_dir):
    #     interval = int(data_size / batch_size) + 1
    #     for epoch in range(num_epochs):
    #
    #         beta_loss_1 = []
    #         beta_loss_2 = []
    #         beta_loss_all = []
    #
    #         batch_shuffled_idx = np.random.permutation(dataset["observations"].shape[0])
    #         for i in tqdm(range(interval)):
    #
    #             start_pt = i * batch_size
    #             end_pt = min((i + 1) * batch_size, dataset["observations"].shape[0])
    #             batch = index_batch(dataset, batch_shuffled_idx[start_pt:end_pt])
    #
    #             # get batch
    #             obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
    #             act_1 = batch['actions']  # batch_size * traj_len * action_dim
    #             obs_2 = batch['observations_2']
    #             act_2 = batch['actions_2']
    #             s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
    #             s_a_2 = np.concatenate([obs_2, act_2], axis=-1)
    #
    #             conditions = [np.all(batch['labels'] == [1, 0], axis=1),
    #                           np.all(batch['labels'] == [0, 1], axis=1),
    #                           np.all(batch['labels'] == [0.5, 0.5], axis=1)]
    #             values = [1, -1, 0]
    #             single_labels = torch.from_numpy(np.select(conditions, values)).float().to(self.device)
    #
    #
    #             pred_1_alpha_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device)) # batch_size * 2
    #             pred_1_alpha = pred_1_alpha_beta[:, 0]
    #             pred_1_beta = pred_1_alpha_beta[:, 1]
    #             # pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #
    #
    #             pred_2_alpha_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #             pred_2_alpha = pred_2_alpha_beta[:, 0]
    #             pred_2_beta = pred_2_alpha_beta[:, 1]
    #             # pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #
    #             # pred_1_alpha, pred_1_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device))
    #             # pred_2_alpha, pred_2_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
    #
    #             pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()
    #             pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
    #             # if equal, then discard
    #             # TODO maybe if equal, then both towards 1 (both preferred)
    #             loss_1 = torch.mean(torch.log(pred_1_preds) * single_labels) - torch.mean(torch.log(pred_2_preds) * single_labels)
    #
    #             # pred_1_preds_alphamask = torch.distributions.Beta(pred_1_alpha.detach(), pred_1_beta).rsample()
    #             # pred_2_preds_alphamask = torch.distributions.Beta(pred_2_alpha.detach(), pred_2_beta).rsample()
    #             #
    #             # pred_1_preds_betamask = torch.distributions.Beta(pred_1_alpha, pred_1_beta.detach()).rsample()
    #             # pred_2_preds_betamask = torch.distributions.Beta(pred_2_alpha, pred_2_beta.detach()).rsample()
    #             #
    #             # alpha_mask = (single_labels == 1).float()
    #             # beta_mask = (single_labels == -1).float()
    #             #
    #             # loss_1_alpha = torch.mean(torch.log(pred_1_preds_betamask) * alpha_mask) + torch.mean(torch.log(pred_2_preds_betamask) * beta_mask)
    #             # loss_1_beta = torch.mean(torch.log(pred_1_preds_alphamask) * beta_mask) + torch.mean(torch.log(pred_2_preds_alphamask) * alpha_mask)
    #             # loss_1 = loss_1_alpha - loss_1_beta
    #             loss_2 = self.kl_regularizer_loss(pred_1_alpha.shape[0], alpha=pred_1_alpha, beta=pred_1_beta) + \
    #                      self.kl_regularizer_loss(pred_1_alpha.shape[0], alpha=pred_2_alpha, beta=pred_2_beta)
    #
    #             #
    #             # target_pred_1_alpha = pred_1_alpha.detach() + (single_labels == 1).float()
    #             # target_pred_1_beta = pred_1_beta.detach() + (single_labels == -1).float()
    #             #
    #             # target_pred_2_alpha = pred_2_alpha.detach() + (single_labels == -1).float()
    #             # target_pred_2_beta = pred_2_beta.detach() + (single_labels == 1).float()
    #             #
    #             # loss_alpha_1 = F.mse_loss(pred_1_alpha, target_pred_1_alpha)
    #             # loss_beta_1 = F.mse_loss(pred_1_beta, target_pred_1_beta)
    #             #
    #             # loss_alpha_2 = F.mse_loss(pred_2_alpha, target_pred_2_alpha)
    #             # loss_beta_2 = F.mse_loss(pred_2_beta, target_pred_2_beta)
    #             #
    #             # loss_1 = loss_alpha_1 + loss_beta_1
    #
    #             beta_loss = -loss_1 + self.beta_coef * loss_2
    #
    #             beta_loss_1.append(-loss_1)
    #             beta_loss_2.append(loss_2)
    #             beta_loss_all.append(beta_loss)
    #
    #             self.opt.zero_grad()
    #             beta_loss.backward()
    #             self.opt.step()
    #
    #         beta_loss_1 = torch.stack(beta_loss_1, dim=0)
    #         beta_loss_2 = torch.stack(beta_loss_2, dim=0)
    #         beta_loss_all = torch.stack(beta_loss_all, dim=0)
    #         print("iteration:", epoch + 1)
    #         print("mean_beta_loss_data:", torch.mean(beta_loss_1).item())
    #         print("mean_beta_loss_kl:", torch.mean(beta_loss_2).item())
    #         print("mean_beta_loss_all:", torch.mean(beta_loss_all).item())
    #
    #         if (epoch+1) % 10 == 0:
    #             tmp_save_dir= save_dir / f'itr_{epoch+1}'
    #             tmp_save_dir.mkdir(parents=True, exist_ok=True)
    #             model_file = tmp_save_dir / 'beta_model.pth'
    #             self.save_model(model_file)

    def fit_data(self, num_epochs, data_size, batch_size, dataset, save_dir):
        interval = int(data_size / batch_size) + 1
        for epoch in range(num_epochs):

            beta_loss_1 = []
            beta_loss_2 = []
            beta_loss_all = []

            batch_shuffled_idx = np.random.permutation(dataset["observations"].shape[0])
            for i in tqdm(range(interval)):

                start_pt = i * batch_size
                end_pt = min((i + 1) * batch_size, dataset["observations"].shape[0])
                batch = index_batch(dataset, batch_shuffled_idx[start_pt:end_pt])

                # get batch
                obs_1 = batch['observations']  # batch_size * traj_len * obs_dim
                act_1 = batch['actions']  # batch_size * traj_len * action_dim
                obs_2 = batch['observations_2']
                act_2 = batch['actions_2']
                s_a_1 = np.concatenate([obs_1, act_1], axis=-1)
                s_a_2 = np.concatenate([obs_2, act_2], axis=-1)

                conditions = [np.all(batch['labels'] == [1, 0], axis=1),
                              np.all(batch['labels'] == [0, 1], axis=1),
                              np.all(batch['labels'] == [0.5, 0.5], axis=1)]
                values = [1, -1, 0]
                single_labels = torch.from_numpy(np.select(conditions, values)).float().to(self.device)


                pred_1_alpha_beta = self.model(torch.from_numpy(s_a_1).float().to(self.device)) # batch_size * 2
                pred_1_alpha = pred_1_alpha_beta[:, 0]
                pred_1_beta = pred_1_alpha_beta[:, 1]
                pred_1_preds = torch.distributions.Beta(pred_1_alpha, pred_1_beta).rsample()


                pred_2_alpha_beta = self.model(torch.from_numpy(s_a_2).float().to(self.device))
                pred_2_alpha = pred_2_alpha_beta[:, 0]
                pred_2_beta = pred_2_alpha_beta[:, 1]
                pred_2_preds = torch.distributions.Beta(pred_2_alpha, pred_2_beta).rsample()
                # if equal, then discard
                # TODO maybe if equal, then both towards 1 (both preferred)
                loss_1 = torch.mean(torch.log(pred_1_preds) * single_labels) - torch.mean(torch.log(pred_2_preds) * single_labels)
                loss_2 = self.kl_regularizer_loss(pred_1_preds.shape[0], alpha=pred_1_alpha, beta=pred_1_beta) + \
                         self.kl_regularizer_loss(pred_2_preds.shape[0], alpha=pred_2_alpha, beta=pred_2_beta)

                beta_loss = -loss_1 + self.beta_coef * loss_2

                beta_loss_1.append(-loss_1)
                beta_loss_2.append(loss_2)
                beta_loss_all.append(beta_loss)

                self.opt.zero_grad()
                beta_loss.backward()
                self.opt.step()

            beta_loss_1 = torch.stack(beta_loss_1, dim=0)
            beta_loss_2 = torch.stack(beta_loss_2, dim=0)
            beta_loss_all = torch.stack(beta_loss_all, dim=0)
            print("iteration:", epoch + 1)
            print("mean_beta_loss_data:", torch.mean(beta_loss_1).item())
            print("mean_beta_loss_kl:", torch.mean(beta_loss_2).item())
            print("mean_beta_loss_all:", torch.mean(beta_loss_all).item())

            if (epoch+1) % 10 == 0:
                tmp_save_dir= save_dir / f'itr_{epoch+1}'
                tmp_save_dir.mkdir(parents=True, exist_ok=True)
                model_file = tmp_save_dir / 'beta_model.pth'
                self.save_model(model_file)

    def fit_data_discrete(self, dataset):

        obs_1 = dataset['observations']  # batch_size * traj_len * obs_dim
        obs_2 = dataset['observations_2']

        conditions = [np.all(dataset['labels'] == [1, 0], axis=1),
                      np.all(dataset['labels'] == [0, 1], axis=1),
                      np.all(dataset['labels'] == [0.5, 0.5], axis=1)]
        values = [1, -1, 0]
        single_labels = torch.from_numpy(np.select(conditions, values)).float().to(self.device)

        discrete_obs_1 = get_discrete_traj(obs_1, dim=2)
        discrete_obs_2 = get_discrete_traj(obs_2, dim=2)

        traj_alpha_beta_dict = init_trajectory_dict(discrete_obs_1 + discrete_obs_2)
        self.traj_alpha_beta_dict = get_trajectory_dict_from_pair(traj_alpha_beta_dict, discrete_obs_1, discrete_obs_2,
                                                                  single_labels)

    def get_alpha_beta_discrete(self, x, dim=2):
        traj = x.cpu().numpy()
        discrete_obs = get_discrete_traj(traj, dim=dim)[0]
        alpha = self.traj_alpha_beta_dict[discrete_obs][0]
        beta = self.traj_alpha_beta_dict[discrete_obs][1]
        return alpha, beta

    def get_normalized_alpha_beta_discrete(self, x, dim=2):
        traj = x.cpu().numpy()
        discrete_obs = get_discrete_traj(traj, dim=dim)[0]
        alpha = self.traj_alpha_beta_dict[discrete_obs][0]
        beta = self.traj_alpha_beta_dict[discrete_obs][1]
        # Normalize alpha and beta
        alpha_normalized = alpha / 10
        beta_normalized = beta / 10
        return alpha_normalized, beta_normalized

    def get_rescaled_alpha_beta_discrete(self, x, dim=2, rescale=10):
        traj = x.cpu().numpy()
        discrete_obs = get_discrete_traj(traj, dim=dim)[0]
        alpha = self.traj_alpha_beta_dict[discrete_obs][0]
        beta = self.traj_alpha_beta_dict[discrete_obs][1]
        # Normalize alpha and beta
        alpha_normalized = np.maximum(alpha / rescale, 1)
        beta_normalized = np.maximum(beta / rescale, 1)

        return alpha_normalized, beta_normalized

    def save_model(self, filepath):
        torch.save(self.state_dict(), filepath)

    def load_model(self, filepath):
        self.load_state_dict(torch.load(filepath, map_location=self.device))

class PrefTransformer1(nn.Module):
    ''' Transformer Structure used in Preference Transformer.
    
    Description:
        This structure holds a causal transformer, which takes in a sequence of observations and actions, 
        and outputs a sequence of latent vectors. Then, pass the latent vectors through self-attention to
        get a weight vector, which is used to weight the latent vectors to get the final preference score.
    
    Args:
        - observation_dim: dimension of observation
        - action_dim: dimension of action
        - max_seq_len: maximum length of sequence
        - d_model: dimension of transformer
        - nhead: number of heads in transformer
        - num_layers: number of layers in transformer
    '''
    def __init__(self,
        observation_dim: int, action_dim: int, 
        max_seq_len: int = 100,
        d_model: int = 256, nhead: int = 4, num_layers: int = 1, 
    ):
        super().__init__()
        self.max_seq_len = max_seq_len
        self.d_model = d_model
        self.pos_emb = SinusoidalPosEmb(d_model)
        
        self.obs_emb = nn.Sequential(
            nn.Linear(observation_dim, d_model),
            nn.LayerNorm(d_model)
        )
        self.act_emb = nn.Sequential(
            nn.Linear(action_dim, d_model),
            nn.LayerNorm(d_model)
        )
        
        self.causual_transformer = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True), 
            num_layers
        )
        self.mask = nn.Transformer.generate_square_subsequent_mask(2*self.max_seq_len)
        
        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.r_proj = nn.Linear(d_model, 1)

    def forward(self, obs: torch.Tensor, act: torch.Tensor):
        if self.mask.device != obs.device: self.mask = self.mask.to(obs.device)
        batch_size, traj_len = obs.shape[:2]
        
        pos = self.pos_emb(
            torch.arange(traj_len, device=obs.device))[None,]
        obs = self.obs_emb(obs) + pos
        act = self.act_emb(act) + pos
        
        x = torch.empty((batch_size, 2*traj_len, self.d_model), device=obs.device)
        x[:, 0::2] = obs
        x[:, 1::2] = act

        x = self.causual_transformer(x, self.mask[:2*traj_len,:2*traj_len])[:, 1::2]
        # x: (batch_size, traj_len, d_model)

        q = self.q_proj(x) # (batch_size, traj_len, d_model)
        k = self.k_proj(x) # (batch_size, traj_len, d_model)
        r = self.r_proj(x) # (batch_size, traj_len, 1)
        
        w = torch.softmax(q@k.permute(0, 2, 1)/np.sqrt(self.d_model), -1).mean(-2)
        # w: (batch_size, traj_len)
        
        z = (w * r.squeeze(-1)) # (batch_size, traj_len)
        
        return torch.tanh(z)


class PrefTransformer2(nn.Module):
    ''' Preference Transformer with no causal mask and no self-attention but one transformer layer to get the weight vector.
    
    Description:
        This structure has no causal mask and no self-attention.
        Instead, it uses one transformer layer to get the weight vector.
        
    Args:
        - observation_dim: dimension of observation
        - action_dim: dimension of action
        - d_model: dimension of transformer
        - nhead: number of heads in transformer
        - num_layers: number of layers in transformer
    '''
    def __init__(self,
        observation_dim: int, action_dim: int, 
        d_model: int, nhead: int, num_layers: int, 
    ):
        super().__init__()
        while num_layers < 2: num_layers += 1
        self.d_model = d_model
        self.pos_emb = SinusoidalPosEmb(d_model)
        self.obs_emb = nn.Sequential(
            nn.Linear(observation_dim, d_model),
            nn.LayerNorm(d_model)
        )
        self.act_emb = nn.Sequential(
            nn.Linear(action_dim, d_model),
            nn.LayerNorm(d_model)
        )
        
        self.transformer = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True), 
            num_layers - 1
        )
        self.value_layer = nn.Sequential(nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True), 1
        ), nn.Linear(d_model, 1))
        self.weight_layer = nn.Sequential(nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True), 1
        ), nn.Linear(d_model, 1))

    def forward(self, obs: torch.Tensor, act: torch.Tensor):
        batch_size, traj_len = obs.shape[:2]
        
        pos = self.pos_emb(
            torch.arange(traj_len, device=obs.device))[None,]
        obs = self.obs_emb(obs) + pos
        act = self.act_emb(act) + pos
        
        x = torch.empty((batch_size, 2*traj_len, self.d_model), device=obs.device)
        x[:, 0::2] = obs
        x[:, 1::2] = act
        
        x = self.transformer(x)[:, 1::2]
        v = self.value_layer(x)
        w = torch.softmax(self.weight_layer(x), 1)
        return (w*v).squeeze(-1)
    

class PrefTransformer3(nn.Module):
    ''' Preference Transformer with no causal mask and no weight vector.
    
    Description:
        This structure has no causal mask and even no weight vector.
        Instead, it directly outputs the preference score.
        
    Args:
        - observation_dim: dimension of observation
        - action_dim: dimension of action
        - d_model: dimension of transformer
        - nhead: number of heads in transformer
        - num_layers: number of layers in transformer
    '''
    def __init__(self,
        observation_dim: int, action_dim: int, 
        d_model: int, nhead: int, num_layers: int, 
    ):
        super().__init__()

        self.d_model = d_model
        self.pos_emb = SinusoidalPosEmb(d_model)
        self.obs_emb = nn.Sequential(
            nn.Linear(observation_dim, d_model),
            nn.LayerNorm(d_model)
        )
        self.act_emb = nn.Sequential(
            nn.Linear(action_dim, d_model),
            nn.LayerNorm(d_model)
        )
        
        self.transformer = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True), 
            num_layers
        )
        self.output_layer = nn.Linear(d_model, 1)

    def forward(self, obs: torch.Tensor, act: torch.Tensor):
        batch_size, traj_len = obs.shape[:2]
        
        pos = self.pos_emb(
            torch.arange(traj_len, device=obs.device))[None,]
        obs = self.obs_emb(obs) + pos
        act = self.act_emb(act) + pos
        
        x = torch.empty((batch_size, 2*traj_len, self.d_model), device=obs.device)
        x[:, 0::2] = obs
        x[:, 1::2] = act
        
        x = self.transformer(x)[:, 1::2]
        return self.output_layer(x).squeeze(-1)


class MLPDiffusion(nn.Module):
    def __init__(
        self,
        obs_dim,
        acs_dim,
        num_step = 200,
        beta_start = 1e-4,
        beta_end = 1e-1,
        traj_len = 200,
        data_dim = 1,
        device="cuda",
    ):
        super().__init__()
        self.obs_dim = obs_dim
        self.acs_dim = acs_dim
        self.num_step = num_step
        self.beta_start = beta_start
        self.beta_end = beta_end
        self.traj_len = traj_len
        self.data_dim = data_dim
        self.device = device

        self.network = nn.Sequential(
            nn.Linear(obs_dim + acs_dim + data_dim + 1, 256),  # input dim: obs_dim + acs_dim + t_dim(=1) + x=data_dim
            nn.LeakyReLU(),
            nn.Linear(256, 256),
            nn.LeakyReLU(),
            nn.Linear(256, 256),
            nn.LeakyReLU(),
            nn.Linear(256, data_dim),
        ).to(self.device)

        self.init_paras()

    def init_paras(self):
        self.beta = torch.linspace(
            start=self.beta_start, end=self.beta_end, steps=self.num_step
        ).view(-1, 1).repeat(1, self.traj_len).to(self.device)
        self.sigma = torch.sqrt(self.beta)
        self.alpha = 1 - self.beta
        self.sqrt_alpha = torch.sqrt(self.alpha)
        self.alpha_bar = torch.cumprod(self.alpha, 0)
        self.sqrt_alpha_bar = torch.sqrt(self.alpha_bar)
        self.one_minus_alpha = 1 - self.alpha
        self.sqrt_one_minus_alpha_bar = torch.sqrt(1 - self.alpha_bar)
        self.multiplier2 = self.one_minus_alpha / self.sqrt_one_minus_alpha_bar
        self.multiplier1 = 1 / self.sqrt_alpha

    def forward(self, obs, action, x, t):  # eps_theta
        if not isinstance(obs, torch.Tensor):
            obs = torch.tensor(obs, dtype=torch.float)
        if not isinstance(obs, torch.Tensor):
            action = torch.tensor(action, dtype=torch.float)
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, dtype=torch.float)
        if not isinstance(t, torch.Tensor):
            t = torch.tensor(t, dtype=torch.float)
        obs = obs.to(self.device)
        action = action.to(self.device)
        t = t.unsqueeze(-1)
        input = torch.cat([obs, action, x, t / self.num_step], dim=2)
        ret = self.network(input)
        return ret

    def reverse_sample(self, obs, action, x_t, t):  # from intermediate noise to data
        mul2_t = self.multiplier2.gather(0, t).unsqueeze(-1)
        mul1_t = self.multiplier1.gather(0, t).unsqueeze(-1)

        eps_theta = self.forward(obs, action, x_t, t)
        mean = mul1_t * (x_t - mul2_t * eps_theta)

        sigma_z = torch.gather(self.sigma, 0, t).unsqueeze(-1) * torch.randn_like(
            x_t, device=self.device
        )

        return mean + sigma_z

    def sample(self, obs, action, during_training=False):  # from pure noise to data (true forward)
        if during_training is False:
            if not isinstance(obs, torch.Tensor):
                if isinstance(obs, list):
                    obs = np.array(obs)
                obs = torch.tensor(obs, dtype=torch.float)
            if not isinstance(action, torch.Tensor):
                action = torch.tensor(action, dtype=torch.int64)
            obs = obs.to(self.device)
            action = action.to(self.device)

        batch_size = obs.shape[0]
        x = torch.randn([batch_size, self.traj_len, self.data_dim], device=self.device)
        for t in reversed(range(self.num_step)):
            x = self.reverse_sample(
                obs,
                action,
                x,
                torch.tensor(t).repeat(batch_size, self.traj_len).to(self.device),
            ).detach()
        return torch.tanh(x)

    def compute_loss(self, obs, action, x_0):  # x_0 is true data, x_t / x is noise
        if not isinstance(obs, torch.Tensor):
            obs = torch.tensor(obs, dtype=torch.long)
        if not isinstance(action, torch.Tensor):
            action = torch.tensor(action, dtype=torch.long)
        batch_size = x_0.shape[0]
        t = torch.randint(0, self.num_step, size=[batch_size], device=self.device).view(-1, 1)
        t = t.repeat(1, self.traj_len)
        sqrt_alpha_bar_t = self.sqrt_alpha_bar.gather(0, t).unsqueeze(-1)
        sqrt_one_minus_alpha_bar_t = self.sqrt_one_minus_alpha_bar.gather(0, t).unsqueeze(-1)
        eps = torch.randn_like(x_0, device=self.device)
        eps_theta = self.forward(
            obs,
            action,
            sqrt_alpha_bar_t * x_0 + eps * sqrt_one_minus_alpha_bar_t,
            t,
        )
        return torch.square(eps - eps_theta).mean()


    # def fit(self, dataset):
    #     loss_list = []
    #     opt = torch.optim.Adam(self.network.parameters(), lr=self.lr)
    #
    #     tmp = []
    #
    #     for _ in tqdm(range(self.max_iter)):
    #
    #         obs, action, reward, next_obs, next_action, done, next_val = dataset.sample(
    #             self.batch_size, self.h_to_learn
    #         )
    #
    #         obs = torch.tensor(obs, dtype=torch.float).to(self.device)
    #         action = torch.tensor(action, dtype=torch.long).to(self.device)
    #         reward = torch.tensor(reward, dtype=torch.float).to(self.device)
    #         next_obs = torch.tensor(next_obs, dtype=torch.float).to(self.device)
    #         next_action = torch.tensor(next_action, dtype=torch.long).to(self.device)
    #         done = torch.tensor(done, dtype=torch.float).to(self.device)
    #         next_val = torch.tensor(next_val, dtype=torch.float).to(self.device)
    #
    #         ret = reward + (1 - done) * next_val
    #         ret = ret.detach().float()
    #
    #         loss = self.compute_loss(
    #             obs,
    #             action,
    #             x_0=ret,
    #         )
    #         loss_list.append(loss.item())
    #         opt.zero_grad()
    #         loss.backward()
    #         opt.step()
    #
    #     return self, loss_list, tmp

LOG_STD_MIN = -5
LOG_STD_MAX = 2

class DistributionalPrefTransformer(nn.Module):

    def __init__(self,
                 observation_dim: int, action_dim: int,
                 max_seq_len: int = 100,
                 d_model: int = 256, nhead: int = 4, num_layers: int = 1,
                 ):
        super().__init__()
        self.max_seq_len = max_seq_len
        self.d_model = d_model
        self.pos_emb = SinusoidalPosEmb(d_model)

        self.obs_emb = nn.Sequential(
            nn.Linear(observation_dim, d_model),
            nn.LayerNorm(d_model)
        )
        self.act_emb = nn.Sequential(
            nn.Linear(action_dim, d_model),
            nn.LayerNorm(d_model)
        )

        self.causual_transformer = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(d_model, nhead, d_model * 4, batch_first=True),
            num_layers
        )
        self.mask = nn.Transformer.generate_square_subsequent_mask(2 * self.max_seq_len)

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.mean_proj = nn.Linear(d_model, 1)
        self.log_std_proj = nn.Linear(d_model, 1)

    def forward(self, obs: torch.Tensor, act: torch.Tensor):
        if self.mask.device != obs.device: self.mask = self.mask.to(obs.device)
        batch_size, traj_len = obs.shape[:2]

        pos = self.pos_emb(
            torch.arange(traj_len, device=obs.device))[None,]
        obs = self.obs_emb(obs) + pos
        act = self.act_emb(act) + pos

        x = torch.empty((batch_size, 2 * traj_len, self.d_model), device=obs.device)
        x[:, 0::2] = obs
        x[:, 1::2] = act

        x = self.causual_transformer(x, self.mask[:2 * traj_len, :2 * traj_len])[:, 1::2]
        # x: (batch_size, traj_len, d_model)

        q = self.q_proj(x)  # (batch_size, traj_len, d_model)
        k = self.k_proj(x)  # (batch_size, traj_len, d_model)
        mean = self.mean_proj(x)  # (batch_size, traj_len, 1)
        log_std = self.log_std_proj(x)  # (batch_size, traj_len, 1)
        # log_std = torch.tanh(log_std)
        # log_std = LOG_STD_MIN + 0.5 * (LOG_STD_MAX - LOG_STD_MIN) * (log_std + 1)
        std = log_std.exp()
        normal = torch.distributions.Normal(mean, std)
        r = normal.rsample()  # (batch_size, traj_len, 1)

        w = torch.softmax(q @ k.permute(0, 2, 1) / np.sqrt(self.d_model), -1).mean(-2)
        # w: (batch_size, traj_len)

        z = (w * r.squeeze(-1))  # (batch_size, traj_len)

        return torch.tanh(z)